"""
Implements generating SpirV code from our bytecode.
"""

import ctypes

from ._generator_base import (
    BaseSpirVGenerator,
    ValueId,
    VariableAccessId,
    WordPlaceholder,
)
from ._coreutils import ShaderError
from . import _spirv_constants as cc
from . import _types
from .stdlib import tex_functions, ext_functions
from .opcodes import OpCodeDefinitions

# todo: build in some checks
# - expect no func or entrypoint inside a func definition
# - expect other opcodes only inside a func definition
# - expect input/output/uniform at the very start (or inside an entrypoint?)


image_formats_that_need_no_ext = {
    cc.ImageFormat_Rgba32f,
    cc.ImageFormat_Rgba16f,
    cc.ImageFormat_Rgba32i,
    cc.ImageFormat_Rgba32ui,
    cc.ImageFormat_Rgba16i,
    cc.ImageFormat_Rgba16ui,
    cc.ImageFormat_Rgba8,
    cc.ImageFormat_Rgba8i,
    cc.ImageFormat_Rgba8ui,
    cc.ImageFormat_Rgba8Snorm,
    cc.ImageFormat_R32f,
    cc.ImageFormat_R32i,
    cc.ImageFormat_R32ui,
}


class Bytecode2SpirVGenerator(OpCodeDefinitions, BaseSpirVGenerator):
    """ A generator that operates on our own well-defined bytecode.

    In essence, this class implements BaseSpirVGenerator by implementing
    the opcode methods of OpCodeDefinitions.
    """

    def show_bytecode(self):
        """ For debugging purposes.
        """
        for x in self._bytecode:
            print(x)

    def _convert(self, bytecode):

        self._bytecode = bytecode
        self._execution_model_flag = None

        self._stack = []
        self._stack_for_phi = {}  # label -> []

        # Track loops. The bottom of the stack is an empty dict, for convenience
        self._loop_stack = [{}]

        # External variables per storage class
        self._input = {}
        self._output = {}
        self._uniform = {}
        self._buffer = {}
        self._sampler = {}
        self._texture = {}
        self._slotmap = {}  # (namespaceidentifier, slot) -> name

        self._decorated_array_types = set()

        # We keep track of sampler for each combination of texture and sampler
        self._texture_samplers = {}

        # Keep track what id a name was saved by. Some need to be stored in variables.
        self._name_ids = {}  # name -> ValueId
        self._name_variables = {}  # name -> VariableAccessId

        # Labels for control flow
        self._labels = {}
        self._root_branch = {"depth": 0, "label": "", "children": ()}
        self._current_branch = self._root_branch

        # Pre-parse the bytecode, to detect what variables names need
        # to be stored in a func-variable, and what can be used via the
        # (faster and easier) SSA form.
        cur_block_label = ""
        saved_in_blocks = {}  # name -> set of labels
        self._need_name_var_save = {}  # label -> set of names
        self._need_name_var_load = {}  # label -> set of names
        for opcode, *args in bytecode:
            if opcode == "co_label":
                cur_block_label = args[0]
            elif opcode == "co_store_name":
                name = args[0]
                saved_in_blocks.setdefault(name, set()).add(cur_block_label)
            elif opcode == "co_load_name":
                name = args[0]
                blocks_where_name_is_saved = saved_in_blocks.get(name, ())
                if cur_block_label in blocks_where_name_is_saved:
                    pass  # no need to use a variable
                else:
                    s = self._need_name_var_load.setdefault(cur_block_label, set())
                    s.add(name)
        # Also mark the blocks where we need to save the variable
        for names in self._need_name_var_load.values():
            for name in names:
                blocks_where_name_is_saved = saved_in_blocks.get(name, ())
                for block_label in blocks_where_name_is_saved:
                    s = self._need_name_var_save.setdefault(block_label, set())
                    s.add(name)

        # Parse
        for opcode, *args in bytecode:
            method = getattr(self, opcode.lower(), None)
            if method is None:
                # pprint_bytecode(self._co)
                raise RuntimeError(f"Cannot parse {opcode} yet.")
            else:
                method(*args)

    def _get_label_id(self, label_value):
        if label_value not in self._labels:
            label_id = self.obtain_id(f"label-{label_value}")
            label_id.resolve(self)
            self._labels[label_value] = label_id
        return self._labels[label_value]

    def co_func(self, name):
        raise ShaderError("No sub-functions yet")

    def co_entrypoint(self, name, shader_type, execution_modes):
        # Special function definition that acts as an entrypoint

        # Get execution_model flag
        modelmap = {
            "compute": cc.ExecutionModel_GLCompute,  # see also ExecutionModel_Kernel
            "vertex": cc.ExecutionModel_Vertex,
            "fragment": cc.ExecutionModel_Fragment,
            "geometry": cc.ExecutionModel_Geometry,
        }
        execution_model_flag = modelmap.get(shader_type.lower(), None)
        self._execution_model_flag = execution_model_flag
        if execution_model_flag is None:
            raise ShaderError(f"Unknown execution model: {shader_type}")

        # Define entry points
        # Note that we must add the ids of all used OpVariables that this entrypoint uses.
        self._entry_point_id = entry_point_id = self.obtain_id(name)
        self.gen_instruction(
            "entry_points", cc.OpEntryPoint, execution_model_flag, entry_point_id, name
        )

        # Define execution modes for each entry point
        assert isinstance(execution_modes, dict)
        self._execution_modes.update(execution_modes)
        modes = self._execution_modes
        if execution_model_flag == cc.ExecutionModel_Fragment:
            if "OriginLowerLeft" not in modes and "OriginUpperLeft" not in modes:
                modes["OriginLowerLeft"] = []
        if execution_model_flag == cc.ExecutionModel_GLCompute:
            if "LocalSize" not in modes:
                modes["LocalSize"] = [1, 1, 1]

        # Declare funcion
        return_type_id = self.obtain_type_id(_types.void)
        func_type_id = self.obtain_id("func_declaration")
        self.gen_instruction(
            "types", cc.OpTypeFunction, func_type_id, return_type_id
        )  # 0 args

        # Start function definition
        func_id = entry_point_id
        func_control = 0  # can specify whether it should inline, etc.
        self.gen_func_instruction(
            cc.OpFunction, return_type_id, func_id, func_control, func_type_id
        )
        self.gen_func_instruction(cc.OpLabel, self.obtain_id("label"))

    def co_func_end(self):
        if self._current_branch is not self._root_branch:
            raise RuntimeError("Function ends with unresolved sub-branches!")
        # End function or entrypoint
        self.gen_func_instruction(cc.OpReturn)
        self.gen_func_instruction(cc.OpFunctionEnd)

    def co_return(self):
        # A discard is only allowed in a fragment shader. Later we might
        # add helper functions, in which case co_return has another
        # function.
        if self._execution_model_flag == cc.ExecutionModel_Fragment:
            self.gen_func_instruction(cc.OpKill)
        else:
            raise ShaderError("Unexpected return/discard")

    def co_call(self, funcname, nargs):

        assert len(self._stack) >= nargs
        args = self._stack[-nargs:]
        self._stack[-nargs:] = []

        assert isinstance(funcname, str)

        if funcname in _types.gpu_types_map:
            # A common type, below we also check for more complex type expressions
            ty = _types.gpu_types_map[funcname]
            self._stack.append(self._typecast(ty, args))
        elif funcname in tex_functions:
            self._texture_call(funcname, args)
        elif funcname in ext_functions:
            self._ext_instruction_call(funcname, args)
        else:
            # Well, it could be a more special type ... try to convert!
            try:
                ty = _types.type_from_name(funcname)
            except Exception:
                ty = None
            if ty is not None:
                self._stack.append(self._typecast(ty, args))
            else:
                raise ShaderError(f"Using invalid function call: {funcname}")

    def _typecast(self, ty, args):
        assert not ty.is_abstract
        if issubclass(ty, _types.Vector):
            result = self._vector_packing(ty, args)
        elif issubclass(ty, _types.Array):
            result = self._array_packing(args)
        elif issubclass(ty, _types.Scalar):
            if len(args) != 1:
                raise ShaderError("Scalar convert needs exactly one argument.")
            result = self._convert_scalar(ty, args[0])
        return result

    def _texture_call(self, funcname, args):
        if funcname in ("imageLoad", "read"):
            tex, coord = args
            self._capabilities.add(cc.Capability_StorageImageReadWithoutFormat)
            tex.depth.value, tex.sampled.value = 0, 2
            if coord.type not in (_types.i32, _types.ivec2, _types.ivec3):
                raise ShaderError("Expected texture coords to be i32, ivec2 or ivec3.")
            vec_sample_type = _types.Vector(4, tex.sample_type)
            result_id, type_id = self.obtain_value(vec_sample_type)
            self.gen_func_instruction(
                cc.OpImageRead, type_id, result_id, tex, coord,
            )
            self._stack.append(result_id)
        elif funcname in ("imageStore", "write"):
            tex, coord, color = args
            self._capabilities.add(cc.Capability_StorageImageWriteWithoutFormat)
            tex.depth.value, tex.sampled.value = 0, 2
            if coord.type not in (_types.i32, _types.ivec2, _types.ivec3):
                raise ShaderError("Expected texture coords to be i32, ivec2 or ivec3.")
            if tex.sample_type is _types.i32 and color.type is not _types.ivec4:
                raise ShaderError(
                    f"Expected texture value to be ivec4, not {color.type}"
                )
            elif tex.sample_type is _types.f32 and color.type is not _types.vec4:
                raise ShaderError(
                    f"Expected texture value to be vec4, not {color.type}"
                )
            self.gen_func_instruction(cc.OpImageWrite, tex, coord, color)
            self._stack.append(None)  # this call returns None, gets popped
        elif funcname == "sample":  # -> from a texture
            tex, sam, coord = args
            tex.depth.value, tex.sampled.value = 0, 1
            sample_type = tex.sample_type
            result_id, type_id = self.obtain_value(_types.Vector(4, sample_type))
            self.gen_func_instruction(
                cc.OpImageSampleExplicitLod,  # or cc.OpImageSampleImplicitLod,
                type_id,
                result_id,
                self.get_texture_sampler(tex, sam),
                coord,
                cc.ImageOperandsMask_MaskNone | cc.ImageOperandsMask_Lod,
                self.obtain_constant(0.0),
            )
            self._stack.append(result_id)
        else:
            raise RuntimeError(f"Unknown texture func {funcname}")

    def _ext_instruction_call(self, funcname, args):
        # An extension instruction call. If there is an info dict for
        # this function name, all args must be float or float-vector,
        # and the result is either the same, or the same as the
        # component type. All ext instructions that do not fall into
        # this category are handled seperately here. These are what we
        # call the "hardcoded" functions in stdlib.py.

        # https://www.khronos.org/registry/spir-v/specs/unified1/GLSL.std.450.html
        set_name = "GLSL.std.450"  # The most common

        info = ext_functions.get(funcname, None)

        if info:
            # One of the many float/vec-float functions that we can handle automatically
            set_name, nr, nargs = info["set_name"], info["nr"], info["nargs"]
            # Check
            ty = args[0].type
            if issubclass(ty, _types.Float):
                pass  # ok
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Float):
                pass  # ok
            else:
                raise ShaderError(f"Arg 0 of {funcname} must be float or float-vector.")
            for i in range(1, len(args)):
                if args[i].type is not ty:
                    raise ShaderError(
                        f"Arg {i} of {funcname} must be float or float-vector."
                    )
            # Get result object
            result_type = info["result_type"]
            if result_type == "same":
                result_type = args[0].type
            elif result_type == "component":
                result_type = args[0].type.subtype
            else:
                result_type = s_types.type_from_name(result_type)
        elif funcname == "abs":
            nargs = 1
            result_type = ty = args[0].type
            if issubclass(ty, _types.Float):
                nr = 4
            elif issubclass(ty, _types.Int):
                nr = 5
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Float):
                nr = 4
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Int):
                nr = 5
            else:
                raise ShaderError("abs() expects (vector of) int or float.")
        elif funcname == "sign":
            nargs = 1
            result_type = ty = args[0].type
            if issubclass(ty, _types.Float):
                nr = 6
            elif issubclass(ty, _types.Int):
                nr = 6
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Float):
                nr = 6
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Int):
                nr = 7
            else:
                raise ShaderError("sign() expects (vector of) int or float.")
        elif funcname == "matrix_inverse":
            nargs = 1
            result_type = ty = args[0].type
            if issubclass(ty, _types.Matrix) and ty.rows == ty.cols:
                nr = 34
            else:
                raise ShaderError("matrix_inverse() expects square matrix.")
        elif funcname == "min":
            nargs = 2
            result_type = ty = args[0].type
            if issubclass(ty, _types.Float):
                nr = 37
            elif issubclass(ty, _types.Int):
                nr = 39
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Float):
                nr = 37
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Int):
                nr = 39
            else:
                raise ShaderError("min() expects (vector of) int or float.")
        elif funcname == "max":
            nargs = 2
            result_type = ty = args[0].type
            if issubclass(ty, _types.Float):
                nr = 40
            elif issubclass(ty, _types.Int):
                nr = 42
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Float):
                nr = 40
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Int):
                nr = 42
            else:
                raise ShaderError("max() expects (vector of) int or float.")
        elif funcname == "clamp":
            nargs = 3
            result_type = ty = args[0].type
            if issubclass(ty, _types.Float):
                nr = 43
            elif issubclass(ty, _types.Int):
                nr = 45
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Float):
                nr = 43
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Int):
                nr = 45
            else:
                raise ShaderError("clamp() expects (vector of) int or float.")
        elif funcname == "mix":
            nargs = 3
            nr = 46
            result_type = ty = args[0].type
            if issubclass(ty, _types.Float):
                pass
            elif issubclass(ty, _types.Vector) and issubclass(ty.subtype, _types.Float):
                if issubclass(args[2].type, _types.Float):
                    # Support mix(x, y, a) with X and Y vectors, and A a float
                    result_id, vector_type_id = self.obtain_value(ty)
                    self.gen_func_instruction(
                        cc.OpCompositeConstruct,
                        vector_type_id,
                        result_id,
                        *[args[2]] * ty.length,
                    )
                    args[2] = result_id
            else:
                raise ShaderError("mix() expects (vector of) float.")
        else:
            raise RuntimeError(f"Unknown extension instruction {funcname}")

        # Check
        if nargs != len(args):
            raise ShaderError(
                f"Ext function {funcname} expects {info['nargs']} args, got {nargs}."
            )

        # Generate instruction
        result_id, type_id = self.obtain_value(result_type)
        instr_set = self.obtain_extended_instruction_set(set_name)
        self.gen_func_instruction(
            cc.OpExtInst, type_id, result_id, instr_set, nr, *args
        )
        self._stack.append(result_id)

    # %% IO

    def co_resource(self, name, kind, slot, typename):

        bindgroup = 0
        if isinstance(slot, (tuple, list)):
            bindgroup, slot = slot

        # --> https://www.khronos.org/opengl/wiki/Layout_Qualifier_(GLSL)

        # todo: should we check if the incoming code has the proper amount of skipping
        # in input and output slots?

        # Triage over input kind
        if kind == "input":
            storage_class, iodict = cc.StorageClass_Input, self._input
            # This is also called "shaderLocation" or "slot" in wgpu
            location_or_binding = cc.Decoration_Location
        elif kind == "output":
            storage_class, iodict = cc.StorageClass_Output, self._output
            location_or_binding = cc.Decoration_Location
        elif kind == "uniform":  # slot == binding
            storage_class, iodict = cc.StorageClass_Uniform, self._uniform
            location_or_binding = cc.Decoration_Binding
        elif kind == "buffer":  # slot == binding
            # note: this should be cc.StorageClass_StorageBuffer in SpirV 1.4+
            storage_class, iodict = cc.StorageClass_Uniform, self._buffer
            location_or_binding = cc.Decoration_Binding
        elif kind == "sampler":
            storage_class, iodict = cc.StorageClass_UniformConstant, self._sampler
            location_or_binding = cc.Decoration_Binding
        elif kind == "texture":
            storage_class, iodict = cc.StorageClass_UniformConstant, self._texture
            location_or_binding = cc.Decoration_Binding
        else:
            raise ShaderError(f"Invalid IO kind {kind}")

        # Check if slot is taken.
        if kind in ("input", "output"):
            # Locations must be unique per kind.
            namespace_id = kind
        else:
            # Bindings must be unique within a bind group.
            namespace_id = "bindgroup-" + str(bindgroup)
        slotmap_key = (namespace_id, slot)
        if slotmap_key in self._slotmap:
            other_name = self._slotmap[slotmap_key]
            raise ShaderError(
                f"The {namespace_id} {slot} for {name} already taken by {other_name}."
            )
        else:
            self._slotmap[slotmap_key] = name

        # Get the root variable
        if kind in ("input", "output"):
            var_name = "var-" + name
            var_type = _types.type_from_name(typename)
            subtypes = None
        elif kind in ("uniform", "buffer"):
            # Block - Consider the variable to be a struct
            var_name = "var-" + name
            var_type = _types.type_from_name(typename)
            # Block needs to be a struct
            if issubclass(var_type, _types.Struct):
                subtypes = None
            else:
                subtypes = {name: var_type}
                var_type = _types.Struct(**subtypes)
                var_name = "var-" + var_type.__name__
        elif kind == "sampler":
            var_name = "var-" + name
            var_type = (cc.OpTypeSampler,)
            subtypes = None
        elif kind == "texture":
            var_name = "var-" + name
            # Get a list of type info parts
            type_info = typename.lower().replace(",", " ").split()
            # Get dimension of the texture, and whether it is arrayed
            arrayed = 0
            if "1d" in type_info or "1d-array" in type_info:
                dim = cc.Dim_Dim1D
                arrayed = 1 if "1d-array" in type_info else 0
                self._capabilities.add(cc.Capability_Image1D)
                # self._capabilities.add(cc.Capability_Sampled1D)
            elif "2d" in type_info or "2d-array" in type_info:
                dim = cc.Dim_Dim2D
                arrayed = 1 if "2d-array" in type_info else 0
            elif "3d" in type_info or "3d-array" in type_info:
                dim = cc.Dim_Dim3D
                arrayed = 1 if "3d-array" in type_info else 0
            elif "cube" in type_info or "cube-array" in type_info:
                dim = cc.Dim_Cube
                arrayed = 1 if "cube-array" in type_info else 0
            else:
                raise ShaderError("Texture type info does not specify dimensionality.")
            # Get format
            fmt = cc.ImageFormat_Unknown
            sample_type = None  # can be set through format or specified explicitly
            for part in type_info:
                if part.startswith("r"):
                    part = part.replace("uint", "ui").replace("sint", "i")
                    part = part.replace("int", "i").replace("float", "f")
                    try:
                        fmt = getattr(cc, "ImageFormat_R" + part[1:])
                    except AttributeError:
                        continue
                    if part.endswith(("f", "norm")):
                        sample_type = _types.f32
                    else:
                        sample_type = _types.i32
                    break
            if fmt and fmt not in image_formats_that_need_no_ext:
                self._capabilities.add(cc.Capability_StorageImageExtendedFormats)
            # Get sample type (type of each of the 4 components when sampling)
            if "i32" in type_info:
                sample_type = _types.i32
            elif "f32" in type_info:
                sample_type = _types.f32
            if sample_type is None:  # note that it can have been set from fmt
                raise ShaderError(
                    "Texture type info does not specify format nor sample type."
                )
            # Get whether the texture is sampled - 0: unknown, 1: sampled, 2: storage
            sampled = WordPlaceholder(0)  # -> to be set later
            # Get depth, ms - 0: no, 1: yes, 2: unknown
            depth = WordPlaceholder(2)
            # Get multisampling
            ms = 1 if "ms" in type_info else 0
            # We now have all the info!
            stype = self.obtain_type_id(sample_type)
            var_type = (cc.OpTypeImage, stype, dim, depth, arrayed, ms, sampled, fmt)
            subtypes = None
        else:
            assert False  # unreachable

        # Create VariableAccessId object
        type_id = self.obtain_type_id(var_type)
        var_access = self.obtain_variable(var_type, storage_class, var_name)
        var_id = var_access.variable

        # On textures, store some more info that we need when sampling
        if kind == "texture":
            var_access.sample_type = sample_type
            var_access.sampled = sampled  # a word placeholder
            var_access.depth = depth  # a word placeholder

        # Dectorate block for uniforms and buffers
        if kind == "uniform":
            assert issubclass(var_type, _types.Struct)
            offset = 0
            for i, key in enumerate(var_type.keys):
                subtype = var_type.get_subtype(key)
                offset += self._annotate_uniform_subtype(type_id, subtype, i, offset)
            self.gen_instruction(
                "annotations", cc.OpDecorate, type_id, cc.Decoration_Block
            )
        elif kind == "buffer":
            # todo: according to docs, in SpirV 1.4+, BufferBlock is deprecated
            # and one should use Block with StorageBuffer. But this crashes.
            # Generate an ArrayStride on the storage buffer array
            if issubclass(var_type, _types.Struct) and len(var_type.keys) == 1:
                array_type = var_type.get_subtype(0)
                if issubclass(array_type, _types.Array):
                    stride = ctypes.sizeof(array_type.subtype._as_ctype())
                    if stride > 0 and array_type not in self._decorated_array_types:
                        self._decorated_array_types.add(array_type)
                        self.gen_instruction(
                            "annotations",
                            cc.OpDecorate,
                            self.obtain_type_id(array_type),
                            cc.Decoration_ArrayStride,
                            stride,
                        )
                    self.gen_instruction(
                        "annotations",
                        cc.OpMemberDecorate,
                        type_id,
                        0,
                        cc.Decoration_Offset,
                        0,
                    )
            self.gen_instruction(
                "annotations", cc.OpDecorate, type_id, cc.Decoration_BufferBlock
            )

        # Define slot of variable
        if kind in ("buffer", "texture", "uniform", "sampler"):
            assert isinstance(slot, int)
            # Default to descriptor set zero
            self.gen_instruction(
                "annotations",
                cc.OpDecorate,
                var_id,
                cc.Decoration_DescriptorSet,
                bindgroup,
            )
            self.gen_instruction(
                "annotations", cc.OpDecorate, var_id, cc.Decoration_Binding, slot
            )
        elif isinstance(slot, int):
            self.gen_instruction(
                "annotations", cc.OpDecorate, var_id, location_or_binding, slot
            )
        elif isinstance(slot, str):
            # Builtin input or output
            if slot == "FragDepth":
                self._execution_modes["DepthReplacing"] = []
            try:
                slot = cc.builtins[slot]
            except KeyError:
                raise ShaderError(f"Not a known builtin io variable: {slot}")
            self.gen_instruction(
                "annotations", cc.OpDecorate, var_id, cc.Decoration_BuiltIn, slot
            )

        # Store internal info to derefererence the variables
        if subtypes is None:
            if name in iodict:
                raise ShaderError(f"{kind} {name} already exists")
            iodict[name] = var_access
        else:
            for i, subname in enumerate(subtypes):
                index_id = self.obtain_constant(i)
                if subname in iodict:
                    raise ShaderError(f"{kind} {subname} already exists")
                iodict[subname] = var_access.index(index_id, i)

    def _annotate_uniform_subtype(self, type_id, subtype, i, offset):
        """ Annotates the given uniform struct subtype and return its size in bytes.
        """
        a = "annotations"
        if issubclass(subtype, _types.Matrix):
            # Stride for col or row depending on what is major
            stride = subtype.rows * ctypes.sizeof(subtype.subtype._ctype)
            self.gen_instruction(
                "annotations", cc.OpMemberDecorate, type_id, i, cc.Decoration_ColMajor,
            )
            self.gen_instruction(
                a, cc.OpMemberDecorate, type_id, i, cc.Decoration_MatrixStride, stride,
            )
            self.gen_instruction(
                a, cc.OpMemberDecorate, type_id, i, cc.Decoration_Offset, offset,
            )
        else:
            self.gen_instruction(
                a, cc.OpMemberDecorate, type_id, i, cc.Decoration_Offset, offset,
            )
        return ctypes.sizeof(subtype._as_ctype())

    def get_texture_sampler(self, texture, sampler):
        """ texture and sampler are bot VariableAccessId.
        """
        key = (id(texture), id(sampler))
        if key not in self._texture_samplers:
            tex_type_id = self.obtain_type_id(texture.type)
            restype = (cc.OpTypeSampledImage, tex_type_id)
            result_id, type_id = self.obtain_value(restype)
            self.gen_func_instruction(
                cc.OpSampledImage, type_id, result_id, texture, sampler
            )
            self._texture_samplers[key] = result_id
        return self._texture_samplers[key]

    # %% Basics

    def co_pop_top(self):
        self._stack.pop()

    def co_dup_top(self):
        ob = self._stack.pop()
        self._stack.append(ob)
        self._stack.append(ob)

    def co_rotate_stack(self, n):
        obs = [self._stack.pop() for i in range(n)]
        obs.append(obs.pop(0))
        obs.reverse()
        self._stack.extend(obs)

    def co_reverse_stack(self, n):
        obs = [self._stack.pop() for i in range(n)]
        self._stack.extend(obs)

    def co_load_name(self, name):
        # store a variable that is used in an inner scope.
        if name in self._name_ids:
            # First check if we need to load the name from a variable. If we do,
            # load it. Thereafter we won't need to load it again (in this block).
            current_label = self._current_branch["label"]
            names_that_need_load = self._need_name_var_load.get(current_label, set())
            if name in names_that_need_load:
                ob = self._name_variables[name].resolve_load(self)
                self._name_ids[name] = ob
                names_that_need_load.discard(name)
            ob = self._name_ids[name]
        elif name in self._input:
            ob = self._input[name]
            assert isinstance(ob, VariableAccessId)
            # We could make it quicker for next time, but we loose the deeper access
            # ob = ob.resolve_load(self)
            # self._name_ids[name] = ob
        elif name in self._output:
            ob = self._output[name]
            assert isinstance(ob, VariableAccessId)
        elif name in self._uniform:
            ob = self._uniform[name]
            assert isinstance(ob, VariableAccessId)
        elif name in self._buffer:
            ob = self._buffer[name]
            assert isinstance(ob, VariableAccessId)
        elif name in self._sampler:
            ob = self._sampler[name]
            assert isinstance(ob, VariableAccessId)
        elif name in self._texture:
            ob = self._texture[name]
            assert isinstance(ob, VariableAccessId)
        else:
            raise ShaderError(f"Using invalid variable: {name}")
        self._stack.append(ob)

    def co_store_name(self, name):
        ob = self._stack.pop()
        if name in self._output:
            ac = self._output[name]
            ac.resolve_store(self, ob)
        elif name in self._buffer:
            ac = self._buffer[name]
            ac.resolve_store(self, ob)
        elif name in self._input:
            raise ShaderError("Cannot store to input")
        elif name in self._uniform:
            raise ShaderError("Cannot store to uniform")

        # Store where the result can now be fetched by name (within this block)
        self._name_ids[name] = ob

    def co_load_index(self):
        index = self._stack.pop()
        container = self._stack.pop()

        # Get type of object and index
        element_type = container.type.subtype
        # assert index.type is int

        if isinstance(container, VariableAccessId):
            result_id = container.index(index)

        elif issubclass(container.type, _types.Array):

            # todo: maybe ... the variable should be created only once ...
            # ... instead of every time it gets indexed
            # Put the array into a variable
            var_access = self.obtain_variable(container.type, cc.StorageClass_Function)
            container_variable = var_access.variable
            var_access.resolve_store(self, container.id)

            # Prepare result id and type
            result_id, result_type_id = self.obtain_value(element_type)

            # Create pointer into the array
            pointer1 = self.obtain_id("pointer")
            pointer2 = self.obtain_id("pointer")
            self.gen_instruction(
                "types",
                cc.OpTypePointer,
                pointer1,
                cc.StorageClass_Function,
                result_type_id,
            )
            self.gen_func_instruction(
                cc.OpInBoundsAccessChain, pointer1, pointer2, container_variable, index
            )

            # Load the element from the array
            self.gen_func_instruction(cc.OpLoad, result_type_id, result_id, pointer2)
        else:
            raise ShaderError("Can only index from Arrays")

        self._stack.append(result_id)

        # OpVectorExtractDynamic: Extract a single, dynamically selected, component of a vector.
        # OpVectorInsertDynamic: Make a copy of a vector, with a single, variably selected, component modified.
        # OpVectorShuffle: Select arbitrary components from two vectors to make a new vector.
        # OpCompositeInsert: Make a copy of a composite object, while modifying one part of it. (updating an element)

    def co_store_index(self):
        index = self._stack.pop()
        ob = self._stack.pop()
        val = self._stack.pop()  # noqa

        if isinstance(ob, VariableAccessId):
            # Create new variable access for this last indexing op
            ac = ob.index(index)
            if val.type is not ac.type:
                raise ShaderError(
                    f"Cannot set-index a {val.type.__name__} object in a {ac.type.__name__} container."
                )
            # Then resolve the chain to a store op
            ac.resolve_store(self, val)
        else:
            raise ShaderError(f"Cannot set-index on {ob}")

    def co_load_attr(self, name):
        ob = self._stack.pop()

        if not isinstance(getattr(ob, "type"), type):
            raise ShaderError("Invalid attribute access")
        elif isinstance(ob, VariableAccessId) and issubclass(ob.type, _types.Struct):
            # Struct attribute access
            if name not in ob.type.keys:
                raise ShaderError(f"Attribute {name} invalid for {ob.type.__name__}.")
            # Create new variable access for this attr op
            index = ob.type.keys.index(name)
            ac = ob.index(self.obtain_constant(index), index)
            self._stack.append(ac)
        elif issubclass(ob.type, _types.Vector):
            indices = []
            # Wr support xyzw, rgba, stpq
            for c in name:
                if c in "xrs":
                    indices.append(0)
                elif c in "ygt":
                    indices.append(1)
                elif c in "zbp":
                    indices.append(2)
                elif c in "waq":
                    indices.append(3)
                else:
                    raise ShaderError(f"Invalid vector attribute {name}")
            if len(indices) == 1:
                if isinstance(ob, VariableAccessId):
                    index_id = self.obtain_constant(indices[0])
                    result_id = ob.index(index_id)
                else:
                    result_id, type_id = self.obtain_value(ob.type.subtype)
                    self.gen_func_instruction(
                        cc.OpCompositeExtract, type_id, result_id, ob, indices[0]
                    )
            else:
                result_type = _types.Vector(len(indices), ob.type.subtype)
                result_id, type_id = self.obtain_value(result_type)
                self.gen_func_instruction(
                    cc.OpVectorShuffle, type_id, result_id, ob, ob, *indices
                )
            self._stack.append(result_id)
        else:
            # todo: not implemented for non VariableAccessId
            raise ShaderError(f"Unsupported attribute access {name}")

    def co_load_constant(self, value):
        id = self.obtain_constant(value)
        self._stack.append(id)
        # Also see OpConstantNull OpConstantSampler OpConstantComposite

    def co_load_array(self, nargs):
        # Literal array
        assert len(self._stack) >= nargs
        args = self._stack[-nargs:]
        self._stack[-nargs:] = []
        result = self._array_packing(args)
        self._stack.append(result)

    # %% Math and more

    def co_unary_op(self, op):

        val1 = self._stack.pop()

        # Get reference types
        type1 = val1.type
        reftype1 = type1
        if issubclass(type1, (_types.Vector, _types.Matrix)):
            reftype1 = type1.subtype
        tn1 = type1.__name__

        # Determine opcode and check types
        if op == "neg":
            if issubclass(reftype1, _types.Float):
                opcode = cc.OpFNegate
            elif issubclass(reftype1, _types.Int):
                opcode = cc.OpSNegate
            else:
                raise ShaderError(f"Cannot {op.upper()} values of type {tn1}.")
        elif op == "not":
            if issubclass(reftype1, _types.boolean):
                opcode = cc.OpLogicalNot
            else:
                raise ShaderError(f"Cannot {op.upper()} values of type {tn1}.")

        # Emit code
        result_id, type_id = self.obtain_value(type1)
        self.gen_func_instruction(opcode, type_id, result_id, val1)
        self._stack.append(result_id)

    def co_binary_op(self, op):

        val2 = self._stack.pop()
        val1 = self._stack.pop()

        # The ids that will be in the instruction, can be reset
        id1, id2 = val1, val2

        # Predefine some types
        # We specify three flavors of div: One that works for both int and float,
        # one that works for float only, and one that works for int only.
        scalar_or_vector = _types.Scalar, _types.Vector
        FOPS = dict(
            add=cc.OpFAdd,
            sub=cc.OpFSub,
            mul=cc.OpFMul,
            fdiv=cc.OpFDiv,
            div=cc.OpFDiv,
            mod=cc.OpFMod,
            rem=cc.OpFRem,
        )
        IOPS = dict(
            add=cc.OpIAdd,
            sub=cc.OpISub,
            mul=cc.OpIMul,
            idiv=cc.OpSDiv,
            div=cc.OpSDiv,
            mod=cc.OpSMod,
            rem=cc.OpSRem,
        )
        LOPS = {"and": cc.OpLogicalAnd, "or": cc.OpLogicalOr}

        # Get reference types
        type1 = val1.type
        reftype1 = type1
        if issubclass(type1, (_types.Vector, _types.Matrix)):
            reftype1 = type1.subtype
        type2 = val2.type
        reftype2 = type2
        if issubclass(type2, (_types.Vector, _types.Matrix)):
            reftype2 = type2.subtype

        tn1 = type1.__name__
        tn2 = type2.__name__

        if reftype1 is not reftype2:
            # Let's start by excluding cases where the subtypes differ.
            raise ShaderError(
                f"Cannot {op.upper()} two values with different (sub)types: {tn1} and {tn2}"
            )

        elif type1 is type2 and issubclass(type1, scalar_or_vector):
            # Types are equal and scalar or vector. Covers a lot of cases.
            result_id, type_id = self.obtain_value(type1)
            if (
                issubclass(type1, _types.Vector)
                and issubclass(reftype1, _types.Float)
                and op == "mmul"
            ):
                opcode = cc.OpDot  # special case
                result_id, type_id = self.obtain_value(type1.subtype)
            elif issubclass(reftype1, _types.Float):
                opcode = FOPS[op]
            elif issubclass(reftype1, _types.Int):
                opcode = IOPS[op]
            elif issubclass(reftype1, _types.boolean):
                opcode = LOPS[op]
            else:
                raise ShaderError(f"Cannot {op.upper()} values of type {tn1}.")

        elif issubclass(type1, _types.Scalar) and issubclass(type2, _types.Vector):
            # Convenience - add/mul vectors with scalars
            if not issubclass(reftype1, _types.Float):
                raise ShaderError(
                    f"Scalar {op.upper()} Vector only supported for float subtype."
                )
            result_id, type_id = self.obtain_value(type2)  # result is vector
            if op == "mul":
                opcode = cc.OpVectorTimesScalar
                id1, id2 = val2, val1  # swap to put vector first
            else:
                opcode = FOPS[op]
                val3 = self._vector_packing(type2, [val1] * type2.length)
                id1, id2 = val1, val3

        elif issubclass(type1, _types.Vector) and issubclass(type2, _types.Scalar):
            # Convenience - add/mul vectors with scalars, opposite order
            if not issubclass(reftype1, _types.Float):
                raise ShaderError(
                    f"Vector {op.upper()} Scalar only supported for float subtype."
                )
            result_id, type_id = self.obtain_value(type1)  # result is vector
            if op == "mul":
                opcode = cc.OpVectorTimesScalar
                id1, id2 = val1, val2
            else:
                opcode = FOPS[op]
                val3 = self._vector_packing(type1, [val2] * type1.length)
                id1, id2 = val1, val3

        elif op != "mul":
            # The remaining cases are all limited to multiplication
            raise ShaderError(f"Cannot {op.upper()} {tn1} and {tn2}, multiply only.")

        elif not issubclass(reftype1, _types.Float):
            # The remaining cases are all limited to float types
            raise ShaderError(f"Cannot {op.upper()} {tn1} and {tn2}, float only.")

        # With that out of the way, the remaining cases are quite short to write.

        elif issubclass(type1, _types.Matrix) and issubclass(type2, _types.Matrix):
            # Multiply two matrices
            if type1.cols != type2.rows:
                raise ShaderError(
                    f"Cannot {op.upper()} two matrices with incompatible shapes."
                )
            type3 = _types.Matrix(type2.cols, type1.rows, type1.subtype)
            result_id, type_id = self.obtain_value(type3)
            opcode = cc.OpMatrixTimesMatrix

        elif issubclass(type1, _types.Matrix) and issubclass(type2, _types.Scalar):
            # Matrix times vector
            result_id, type_id = self.obtain_value(type1)  # Result is a matrix
            opcode = cc.OpMatrixTimesScalar
            id1, id2 = val1, val2

        elif issubclass(type1, _types.Matrix) and issubclass(type2, _types.Scalar):
            # Matrix times vector, opposite order
            result_id, type_id = self.obtain_value(type2)  # Result is a matrix
            opcode = cc.OpMatrixTimesScalar
            id1, id2 = val2, val1  # reverse

        elif issubclass(type1, _types.Matrix) and issubclass(type2, _types.Vector):
            # Matrix times Vector
            if type2.length != type1.cols:
                raise ShaderError(f"Incompatible shape for {tn1} x {tn2}")
            type3 = _types.Vector(type1.rows, type1.subtype)
            result_id, type_id = self.obtain_value(type3)
            opcode = cc.OpMatrixTimesVector

        elif issubclass(type1, _types.Vector) and issubclass(type2, _types.Matrix):
            # Vector times Matrix
            if type1.length != type2.rows:
                raise ShaderError(f"Incompatible shape for {tn1} x {tn2}")
            type3 = _types.Vector(type2.cols, type2.subtype)
            result_id, type_id = self.obtain_value(type3)
            opcode = cc.OpVectorTimesMatrix

        else:
            raise ShaderError(f"Cannot {op.upper()} values of {tn1} and {tn2}.")

        self.gen_func_instruction(opcode, type_id, result_id, id1, id2)
        self._stack.append(result_id)

    def co_compare(self, cmp):
        val2 = self._stack.pop()
        val1 = self._stack.pop()

        # Get reference type
        if val1.type is not val2.type:
            raise ShaderError("Cannot compare values that do not have the same type.")
        if issubclass(val1.type, _types.Vector):
            reftype = val1.type.subtype
            result_type = _types.Vector(val1.type.length, _types.boolean)
        else:
            reftype = val1.type
            result_type = _types.boolean

        # Get what kind of comparison to do
        opname_suffix_map = {
            "<": "LessThan",
            "<=": "LessThanEqual",
            "==": "Equal",
            "!=": "NotEqual",
            ">": "GreaterThan",
            ">=": "GreaterThanEqual",
        }
        opname_suffix = opname_suffix_map[cmp]

        # Get the actual opcode
        if issubclass(reftype, _types.Float):
            opcode = getattr(cc, "OpFOrd" + opname_suffix)
        elif issubclass(reftype, _types.Int):
            prefix = "OpS" if "Than" in opname_suffix else "OpI"
            opcode = getattr(cc, prefix + opname_suffix)
        else:
            raise ShaderError(f"Cannot compare values of {val1.type.__name__}.")

        # Generate instruction
        result_id, type_id = self.obtain_value(result_type)
        self.gen_func_instruction(opcode, type_id, result_id, val1, val2)
        self._stack.append(result_id)

    # %% Control flow

    def _before_moving_out_of_a_block(self):
        label = self._current_branch["label"]
        self._store_variables_for_other_blocks(label)
        self._store_stack_for_phi_op(label)

    def _store_variables_for_other_blocks(self, label):
        # Before we wrap up the current block ...
        # We may need to store variable to use in other blocks
        # Sort the names to obtain consistent bytecode.
        for name in sorted(self._need_name_var_save.get(label, ())):
            ob = self._name_ids[name]  # Get the last value
            if name not in self._name_variables:
                self._name_variables[name] = self.obtain_variable(
                    ob.type, cc.StorageClass_Function, name
                )
            var_access = self._name_variables[name]
            if ob.type is not var_access.type:
                raise ShaderError(
                    f"Name {name} used for different types {ob.type} and {var_access.type}"
                )
            var_access.resolve_store(self, ob)

    def _store_stack_for_phi_op(self, label):
        # Sometimes, when we jump out of a block, the stack is not
        # empty. When this happens the stack value is to be picked up
        # by a OpPhi, e.g. with ternary operations (.. if .. else ..).
        # We move the remaining item off the stack and push it on a
        # special dict for the phi op.
        if self._stack:
            self._stack_for_phi.setdefault(label, []).extend(self._stack)
            self._stack = []

    def _collect_stack_from_previous_block(self, prev_label):
        # Take over the remaining stack from the previous block (if this is not a merge block)
        if prev_label in self._stack_for_phi:
            self._stack.extend(self._stack_for_phi[prev_label])

    def _collect_stack_from_previous_blocks(self, *prev_labels):
        # Insert a phi op that selects a stack value based on the branch we came from
        assert len(prev_labels) >= 2
        if any(label in self._stack_for_phi for label in prev_labels):
            assert all(label in self._stack_for_phi for label in prev_labels)
            # All but the last item on the sub-stacks must be equal
            substack = self._stack_for_phi[prev_labels[0]][:-1]
            for label in prev_labels:
                substack2 = self._stack_for_phi[label][:-1]
                assert len(substack) == len(substack2)
                assert all(ob is ob2 for ob, ob2 in zip(substack, substack2))
            # We can simply copy this substack
            self._stack.extend(substack)
            # But for the last item we need a Phi op ...
            top_obs = [self._stack_for_phi[x][-1] for x in prev_labels]
            types = [ob.type for ob in top_obs]
            type = top_obs[0].type
            for ob in top_obs:
                assert (
                    ob.type is type
                ), f"Phi stack has objects with different types: {types}"
            result_id, type_id = self.obtain_value(type)
            phi_op = [cc.OpPhi, type_id, result_id]
            for label, ob in zip(prev_labels, top_obs):
                phi_op.extend([ob, self._get_label_id(label)])
            self.gen_func_instruction(*phi_op)
            self._stack.append(result_id)

    def co_label(self, label):
        # We enter a new block. This is where we resolve branch pairs that both
        # jumped to this block. If there are multiple such pairs, we need to
        # introduce extra blocks, because at each label we can merge at most
        # one branch pair (matching an OpSelectionMerge).

        # Details: https://www.khronos.org/registry/spir-v/specs/unified1/SPIRV.html#_a_id_structuredcontrolflow_a_structured_control_flow
        # * The merge block declared by a header block cannot be a merge
        #   block declared by any other header block.
        # * Each header block must strictly dominate its merge block,
        #   unless the merge block is unreachable in the CFG.
        # * A certain block A dominates a block B, where A and B are in the
        #   same function, if every path from the function’s entry point to
        #   block B includes block A. A strictly dominates B only if A
        #   dominates B and A and B are different blocks.
        # * In other words: every pair of diverging branches must merge at a
        #   unique block (i.e. label), and all control flow going through that
        #   block must also pass through the block where the branch-pair
        #   diverged (i.e.the header block).

        new_label = label

        # Get what loop we're in. If this label closes/merges the loop, pop it.
        loop_info = self._loop_stack[-1]
        if loop_info.get("merge_label") == new_label:
            self._loop_stack.pop(-1)

        # Create a mapping to obtain parents of items in the CFG.
        # We don't use a 'parent' attribute on the items, because that makes
        # these dicts really hard to read during debugging.
        parents = {}

        # First we collect the leaf branches that have jumped to this block
        def _collect_leaf_branches(branch):
            if branch["children"]:
                c1, c2 = branch["children"]
                parents[id(c1)] = parents[id(c2)] = branch
                return _collect_leaf_branches(c1) + _collect_leaf_branches(c2)
            elif branch["label"] == new_label:
                return [branch]
            else:
                return []

        leaf_branches = _collect_leaf_branches(self._root_branch)

        # Then we merge branches, one by one, deeper ones first
        def _select_branch_to_merge():
            for branch in leaf_branches:
                if branch["depth"] > 0:
                    parent = parents[id(branch)]
                    siblings = parent["children"]
                    if siblings[0]["label"] == siblings[1]["label"]:
                        # Merging what started at co_branch_conditional
                        return parent
                    elif parent is loop_info.get("branch", None):
                        # This is the main loop branch. Merge if we reached the merge label.
                        # Note that one child branch is pointing to the continue label, not to here.
                        # Don't AND this sub-if, the above elif excludes cases for the next elif.
                        if loop_info.get("merge_label") == new_label:
                            parent["children"] = tuple(
                                c for c in siblings if c in leaf_branches
                            )
                            return parent
                    elif branch["label"] == new_label:
                        # This could be a break
                        sibling_branch = siblings[int(branch is siblings[0])]
                        if sibling_branch["label"] == loop_info.get("merge_label"):
                            parent["children"] = tuple(
                                c for c in siblings if c in leaf_branches
                            )
                            return parent

        branches2merge = []
        while True:
            leaf_branches.sort(key=lambda b: -b["depth"])
            new_leaf = _select_branch_to_merge()
            if not new_leaf:
                break
            new_leaf["label"] = new_leaf["children"][0]["label"]
            # new_leaf["children"] = () -> do further down, we need that info
            for child in new_leaf["children"]:
                if child in leaf_branches:
                    leaf_branches.remove(child)
            leaf_branches.append(new_leaf)
            branches2merge.append(new_leaf)

        # If all is well, we're now left with a single branch. Make it the current.
        if len(leaf_branches) != 1:
            raise ShaderError(
                f"New block ({new_label}) should start with 1 unmerged branches, got {[b['label'] for b in leaf_branches]}"
            )
        self._current_branch = leaf_branches[0]

        # Need an extra hop?
        exta_hop = 0
        if new_label == self._loop_stack[-1].get("continue_label"):
            exta_hop = 1

        # Get hop labels for all the merges.
        hop_labels = [new_label]
        while len(hop_labels) < len(branches2merge) + exta_hop:
            hop_labels.insert(-1, f"{new_label}-hop-{len(hop_labels)}")

        # Emit the code for the extra blocks, and update previous jump ids.
        hop_label = None
        for i, branch in enumerate(branches2merge):
            # Mark the start of a new block
            hop_label = hop_labels[i]
            label_id = self._get_label_id(hop_label)
            self.gen_func_instruction(cc.OpLabel, label_id)
            # Update placeholders
            for child in branch["children"]:
                child["branch_label_placeholder"].value = label_id.id
            branch["merge_label_placeholder"].value = label_id.id
            # We may need to insert a Phi op
            if len(branch["children"]) == 2:
                self._collect_stack_from_previous_blocks(
                    *[b["prev_label"] for b in branch["children"]]
                )
            # Do we need to hop to the next block?
            if hop_label is not new_label:
                branch_label_placeholder = WordPlaceholder(
                    self._get_label_id(hop_labels[i + 1]).id
                )
                branch["branch_label_placeholder"] = branch_label_placeholder
                branch["prev_label"] = hop_label
                self.gen_func_instruction(cc.OpBranch, branch_label_placeholder)
            if i < len(branches2merge) - 1:
                self._store_stack_for_phi_op(hop_label)

        # If we did not create the main label yet, create it now
        if hop_label != new_label:
            self.gen_func_instruction(cc.OpLabel, self._get_label_id(new_label))

        # Collect stack from previous block
        if not branches2merge:
            self._collect_stack_from_previous_block(self._current_branch["prev_label"])

        # Clean up
        for branch in branches2merge:
            branch["children"] = ()

    def co_branch(self, label):
        # Before we leave this block ...
        self._before_moving_out_of_a_block()
        # Initialize branch label. We use a placeholder because we may introduce
        # additional labels for merges, so the id may need to change.
        branch_label = self._get_label_id(label)
        branch_label_placeholder = WordPlaceholder(branch_label.id)
        # Update the label for the currently running branch
        self._current_branch["prev_label"] = self._current_branch["label"]
        self._current_branch["label"] = label
        # Also update the label placeholder for the last jump
        self._current_branch["branch_label_placeholder"] = branch_label_placeholder
        # Mark the end of the block (will be set again at co_label)
        self._current_branch = None
        # Generate instruction, but not if the last instruction already marked
        # the end of a block.
        if self._sections["functions"][-1][0] not in (cc.OpKill,):
            self.gen_func_instruction(cc.OpBranch, branch_label_placeholder)

    def co_branch_conditional(self, true_label, false_label):
        condition = self._stack.pop()
        current_label = self._current_branch["label"]
        # Before we leave this block ...
        self._before_moving_out_of_a_block()
        # Setup tracing for the two new branches. SpirV wants to know
        # beforehand where the two branches meet, so we will need to
        # update the OpSelectionMerge instruction when we find the merge
        # point in co_label.
        branch1_label = WordPlaceholder(self._get_label_id(true_label).id)
        branch2_label = WordPlaceholder(self._get_label_id(false_label).id)
        new_branch1 = {
            "depth": self._current_branch["depth"] + 1,
            "children": (),
            "label": true_label,
            "prev_label": current_label,
            "branch_label_placeholder": branch1_label,
        }
        new_branch2 = {
            "depth": self._current_branch["depth"] + 1,
            "children": (),
            "label": false_label,
            "prev_label": current_label,
            "branch_label_placeholder": branch2_label,
        }
        self._current_branch["children"] = new_branch1, new_branch2
        # Introduce OpSelectionMerge, unless we've already emitted OpLoopMerge
        if self._loop_stack[-1].get("branch") is not self._current_branch:
            merge_label = WordPlaceholder(0)
            self._current_branch["merge_label_placeholder"] = merge_label
            self.gen_func_instruction(cc.OpSelectionMerge, merge_label, 0)
        # Generate the branch instruction
        self.gen_func_instruction(
            cc.OpBranchConditional, condition, branch1_label, branch2_label,
        )

    def co_branch_loop(self, iter_label, continue_label, merge_label):
        # Before we leave this block ...
        self._before_moving_out_of_a_block()
        # Get id's
        iter_id = self._get_label_id(iter_label)
        continue_id = self._get_label_id(continue_label)
        merge_id = self._get_label_id(merge_label)
        # Generate loop merge instruction
        # note: here we can specify request for unroll, min/max iters (1.4+) etc.
        merge_placeholder = WordPlaceholder(merge_id.id)
        self._current_branch["merge_label_placeholder"] = merge_placeholder
        self.gen_func_instruction(cc.OpLoopMerge, merge_placeholder, continue_id, 0)
        # Mark the current branch as a loop, ending at merge_label
        loop_info = {
            "merge_label": merge_label,
            "iter_label": iter_label,
            "continue_label": continue_label,
            "branch": self._current_branch,
        }
        self._loop_stack.append(loop_info)

        # The rest is similar to co_branch()
        branch_label_placeholder = WordPlaceholder(iter_id.id)
        # Update the label for the currently running branch
        self._current_branch["prev_label"] = self._current_branch["label"]
        self._current_branch["label"] = iter_label
        # Also update the label placeholder for the last jump
        self._current_branch["branch_label_placeholder"] = branch_label_placeholder
        # Mark the end of the block (will be set again at co_label)
        self._current_branch = None
        self.gen_func_instruction(cc.OpBranch, iter_id)

    def co_select(self):
        val2 = self._stack.pop()
        val1 = self._stack.pop()
        condition = self._stack.pop()
        if val1.type is not val2.type:
            raise ShaderError(
                "Incompatible types in op_selec: {val1.type.__name__} and {val2.type.__name__}"
            )
        result_id, type_id = self.obtain_value(val1.type)
        self.gen_func_instruction(
            cc.OpSelect, type_id, result_id, condition, val1, val2
        )
        self._stack.append(result_id)

    # %% Helper methods

    def _convert_scalar(self, out_type, arg):
        return self._convert_scalar_or_vector(out_type, out_type, arg, arg.type)

    def _convert_numeric_vector(self, out_type, arg):
        if not (
            issubclass(arg.type, _types.Vector) and arg.type.length == out_type.length
        ):
            raise ShaderError("Vector conversion needs vectors of equal length.")
        return self._convert_scalar_or_vector(
            out_type, out_type.subtype, arg, arg.type.subtype
        )

    def _convert_scalar_or_vector(self, out_type, out_el_type, arg, arg_el_type):

        # This function only works for vectors for numeric types (no bools)
        if out_type is not out_el_type:
            assert issubclass(out_el_type, _types.Numeric) and issubclass(
                arg_el_type, _types.Numeric
            )

        # Is a conversion actually needed?
        if arg.type is out_type:
            return arg

        # Otherwise we need a new value
        result_id, type_id = self.obtain_value(out_type)

        argtname = arg_el_type.__name__
        outtname = out_el_type.__name__

        if issubclass(out_el_type, _types.Float):
            if issubclass(arg_el_type, _types.Float):
                self.gen_func_instruction(cc.OpFConvert, type_id, result_id, arg)
            elif issubclass(arg_el_type, _types.Int):
                op = cc.OpConvertUToF if argtname.startswith("u") else cc.OpConvertSToF
                self.gen_func_instruction(op, type_id, result_id, arg)
            elif issubclass(arg_el_type, _types.boolean):
                zero = self.obtain_constant(0.0, out_el_type)
                one = self.obtain_constant(1.0, out_el_type)
                self.gen_func_instruction(
                    cc.OpSelect, type_id, result_id, arg, one, zero
                )
            else:
                raise ShaderError(f"Cannot convert to float: {arg.type}")

        elif issubclass(out_el_type, _types.Int):
            if issubclass(arg_el_type, _types.Float):
                op = cc.OpConvertFToU if outtname.startswith("u") else cc.OpConvertFToS
                self.gen_func_instruction(cc.OpConvertFToS, type_id, result_id, arg)
            elif issubclass(arg_el_type, _types.Int):
                op = cc.OpUConvert if outtname.startswith("u") else cc.OpSConvert
                self.gen_func_instruction(cc.OpSConvert, type_id, result_id, arg)
            elif issubclass(arg_el_type, _types.boolean):
                zero = self.obtain_constant(0, out_type)
                one = self.obtain_constant(1, out_type)
                self.gen_func_instruction(
                    cc.OpSelect, type_id, result_id, arg, one, zero
                )
            else:
                raise ShaderError(f"Cannot convert to int: {arg.type}")

        elif issubclass(out_el_type, _types.boolean):
            if issubclass(arg_el_type, _types.Float):
                zero = self.obtain_constant(0.0, arg_el_type)
                self.gen_func_instruction(
                    cc.OpFOrdNotEqual, type_id, result_id, arg, zero
                )
            elif issubclass(arg_el_type, _types.Int):
                zero = self.obtain_constant(0, arg_el_type)
                self.gen_func_instruction(cc.OpINotEqual, type_id, result_id, arg, zero)
            elif issubclass(arg_el_type, _types.boolean):
                return arg  # actually covered above
            else:
                raise ShaderError(f"Cannot convert to bool: {arg.type}")
        else:
            raise ShaderError(f"Cannot convert to {out_type}")

        return result_id

    def _vector_packing(self, vector_type, args):

        # Vector conversion of numeric types is easier
        if (
            len(args) == 1
            and issubclass(vector_type.subtype, _types.Numeric)
            and issubclass(args[0].type.subtype, _types.Numeric)
        ):
            return self._convert_numeric_vector(vector_type, args[0])

        n, t = vector_type.length, vector_type.subtype  # noqa
        composite_ids = []

        # Deconstruct
        for arg in args:
            if not isinstance(arg, ValueId):
                raise RuntimeError("Expected a SpirV object")
            if issubclass(arg.type, _types.Scalar):
                comp_id = arg
                if arg.type is not t:
                    comp_id = self._convert_scalar(t, arg)
                composite_ids.append(comp_id)
            elif issubclass(arg.type, _types.Vector):
                # todo: a contiguous subset of the scalars consumed can be represented by a vector operand instead!
                # -> I think this means we can simply do composite_ids.append(arg)
                for i in range(arg.type.length):
                    comp_id, comp_type_id = self.obtain_value(arg.type.subtype)
                    self.gen_func_instruction(
                        cc.OpCompositeExtract, comp_type_id, comp_id, arg, i
                    )
                    if arg.type.subtype is not t:
                        comp_id = self._convert_scalar(t, comp_id)
                    composite_ids.append(comp_id)
            else:
                raise ShaderError(f"Invalid type to compose vector: {arg.type}")

        # Check the length
        if len(composite_ids) != n:
            raise ShaderError(
                f"{vector_type} did not expect {len(composite_ids)} elements"
            )

        assert (
            len(composite_ids) >= 2
        ), "When constructing a vector, there must be at least two Constituent operands."

        # Construct
        result_id, vector_type_id = self.obtain_value(vector_type)
        self.gen_func_instruction(
            cc.OpCompositeConstruct, vector_type_id, result_id, *composite_ids
        )
        # todo: or OpConstantComposite
        return result_id

    def _array_packing(self, args):
        n = len(args)
        if n == 0:
            raise ShaderError("No support for zero-sized arrays.")

        # Check that all args have the same type
        element_type = args[0].type
        composite_ids = args
        for arg in args:
            assert arg.type is element_type, "array type mismatch"

        # Create array class
        array_type = _types.Array(n, element_type)

        result_id, type_id = self.obtain_value(array_type)
        self.gen_func_instruction(
            cc.OpCompositeConstruct, type_id, result_id, *composite_ids
        )
        # todo: or OpConstantComposite

        return result_id



import math
import inspect
from dis import dis as pprint_bytecode
from dis import cmp_op

from ._coreutils import ShaderError
from ._module import ShaderModule
from .opcodes import OpCodeDefinitions as op
from ._dis import dis
from ._types import gpu_types_map
from .stdlib import __all__ as stdlib_func_names


EXTENDED_ARG = dis.opmap["EXTENDED_ARG"]


def python2shader(func):
    """ Convert a Python function to a ShaderModule object.

    Takes the bytecode of the given function and converts it to our
    internal bytecode. From there it can be converted to binary SpirV.
    All in dependency-free pure Python.
    """

    if not inspect.isfunction(func):
        raise TypeError("python2shader expects a Python function.")

    # Detect shader type
    possible_types = "vertex", "fragment", "compute"
    shader_types = [t for t in possible_types if t in func.__name__.lower()]
    if len(shader_types) == 1:
        shader_type = shader_types[0]
    elif len(shader_types) == 0:
        raise NameError(
            "Shader entrypoint must contain 'vertex', 'fragment' or 'compute' to specify shader type."
        )
    else:
        raise NameError("Ambiguous function name: is it a vert, frag or comp shader?")

    # Convert to bytecode
    converter = PyBytecode2Bytecode()
    converter.convert(func, shader_type)
    bytecode = converter.dump()

    return ShaderModule(func, bytecode, f"shader from {func.__name__}")


class PyBytecode2Bytecode:
    """ Convert Python bytecode to our own well-defined bytecode.
    Python bytecode depends on other variables on the code object, and differs
    between Python functions. This class converts this, so that the next step
    of code generation becomes simpler.
    """

    def show_bytecode(self):
        """ For debugging purposes.
        """
        pprint_bytecode(self._co)

    def convert(self, py_func, shader_type):

        # Attributes of code objects: co_code, co_name, co_filename, co_firstlineno,
        # co_argcount, co_kwonlyargcount, co_nlocals, co_consts, co_varnames,
        # co_names, co_cellvars, co_freevars, co_stacksize, co_flags, co_lnotab
        # -> co_lnotab  is line number table
        #    https://svn.python.org/projects/python/branches/pep-0384/Objects/lnotab_notes.txt
        self._py_func = py_func
        self._co = self._py_func.__code__
        self._py_bytecode = self._co.co_code

        self._opcodes = []  # The resulting "bytecode"

        self._input = {}
        self._output = {}
        self._uniform = {}
        self._buffer = {}
        self._texture = {}
        self._sampler = {}

        # Keep track of labels
        self._labels = {}

        # Protected labels wont automatically generate a co_label,
        # and cannot be resolved if block is empty
        self._protected_labels = set()

        # Bytecode is a stack machine.
        self._stack = []

        # Collect info about loop locations beforehand
        self._loops_to_handle = self._pre_detect_loops()

        # The loop_info objects are popped from the above lists and put on this stack
        self._loop_stack = [{}]  # prepend empty dict to be able to do get()

        # todo: allow user to specify name otherwise?
        entrypoint_name = "main"  # py_func.__name__
        self.emit(op.co_entrypoint, entrypoint_name, shader_type, {})

        KINDMAP = {
            "input": self._input,
            "output": self._output,
            "uniform": self._uniform,
            "buffer": self._buffer,
            "sampler": self._sampler,
            "texture": self._texture,
        }

        defaults = list(py_func.__defaults__ or [])
        while len(defaults) < py_func.__code__.co_argcount:
            defaults.insert(0, None)

        # Parse function inputs
        for i in range(py_func.__code__.co_argcount):
            # Get name and resource object
            argname = py_func.__code__.co_varnames[i]
            resource = py_func.__annotations__.get(argname, None)
            if resource is None:
                resource = defaults[i]
            if resource is None:
                raise TypeError(
                    f"pyshader arg {argname} needs type info either as default value or annotation."
                )
            elif isinstance(resource, tuple) and len(resource) == 3:
                kind, slot, subtype = resource
                assert isinstance(kind, str)
                assert isinstance(slot, (int, str, tuple))
                assert isinstance(subtype, (type, str))
                slot = list(slot) if isinstance(slot, tuple) else slot  # json
            else:
                raise TypeError(
                    f"pyshader arg {argname} type info must be a 3-tuple, not {type(resource)}."
                )
            kind = kind.lower()
            subtype = subtype.__name__ if isinstance(subtype, type) else subtype
            # Get dict to store ref in
            try:
                resource_dict = KINDMAP[kind]
            except KeyError:
                raise TypeError(
                    f"pyshader arg {argname} has unknown resource kind '{kind}')."
                )
            # Emit and store in our dict
            self.emit(op.co_resource, kind + "." + argname, kind, slot, subtype)
            resource_dict[argname] = subtype

        self._convert()
        self.emit(op.co_func_end)

    def _stack_pop(self, allow_global=False):
        ob = self._stack.pop()
        if not allow_global:
            if isinstance(ob, str) and ob.startswith("."):
                raise ShaderError(f"Invalid use of (global) {ob[1:]}")
        return ob

    def emit(self, opcode, *args):
        if callable(opcode):
            fcode = opcode.__code__
            opcode = fcode.co_name  # a method of OpCodeDefinitions class
            argnames = [fcode.co_varnames[i] for i in range(fcode.co_argcount)][1:]
            if len(args) != len(argnames):
                raise RuntimeError(
                    f"Got {len(args)} args for {opcode}({', '.join(argnames)})"
                )

        if opcode == "co_branch":
            assert not self._opcodes[-1][0].startswith("co_branch")
        self._opcodes.append((opcode, *args))

    def dump(self):
        return self._opcodes

    def _convert(self):

        self._pointer = 0
        while self._pointer < len(self._py_bytecode):
            if (
                self._loops_to_handle
                and self._pointer == self._loops_to_handle[0]["start"]
            ):
                self._start_loop(self._loops_to_handle.pop(0))
            elif self._pointer == self._loop_stack[-1].get("end"):
                self._end_loop()
            elif (
                self._pointer in self._labels
                and self._pointer not in self._protected_labels
            ):
                label = self._labels[self._pointer]
                last_opcode = self._opcodes[-1][0]
                if last_opcode not in (
                    "co_branch",
                    "co_branch_conditional",
                    "co_branch_loop",
                ):
                    self.emit(op.co_branch, label)
                self.emit(op.co_label, label)
            opname, arg = self._next()
            method_name = "_op_" + opname.lower()
            method = getattr(self, method_name, None)
            if method is None:
                pprint_bytecode(self._co)
                raise RuntimeError(
                    f"Cannot parse py's {opname} yet (no {method_name}())."
                )
            else:
                method(arg)

        # Some post-processing (order is important)
        self._fix_empty_blocks()
        self._fix_or_control_flow()
        self._fix_consistent_labels()

        # Note: at some point we tried to detect ternary ops (xx if yy else zz)
        # and resolved them into op_select. This detection relied on the fact that
        # a ternary op leaves an item at the stack in both its branches.
        # However, this can also happen in: a = b + (c if d else e)
        # In this statement b is put on the stack before entering the ternary,
        # so we'd detect that as part of a ternary. Maybe this can be detected
        # too, but things get complex quickly, and I did not feel confident in
        # this approach anymore. We could consider giving at another shot later,
        # if it matters significantly for performance.

    def _pre_detect_loops(self):

        # Loops can be detected by a jump that goes backwards in the bytecode.
        # We have to examine the bytecode to find the loop structure, and this
        # consists mostly of looking at jumps, so we first detect all jumps.

        # Collect jumps in the bytecode
        jumps = {}
        jump_ops = (
            "JUMP_ABSOLUTE",
            "JUMP_FORWARD",
            "POP_JUMP_IF_FALSE",
            "POP_JUMP_IF_TRUE",
            "JUMP_IF_FALSE_OR_POP",
            "JUMP_IF_TRUE_OR_POP",
        )

        self._pointer = 0
        while self._pointer < len(self._py_bytecode):
            i = self._pointer
            opname, arg = self._next()
            if "JUMP" in opname:
                assert opname in jump_ops
                jumps[i] = (i + 2 + arg) if opname == "JUMP_FORWARD" else arg

        # Look for loop starts
        loop_starts = []
        for i, target in jumps.items():
            if target < i and target not in loop_starts:
                loop_starts.append(target)

        # Sort the starts: this is the order in which thery are encountered!
        loop_starts.sort()

        # Return list of loop_info objects
        loop_infos = []
        for i in range(len(loop_starts)):
            loop_info = self._pre_detect_loop(jumps, loop_infos, loop_starts[i])
            loop_infos.append(loop_info)
        return loop_infos

    def _pre_detect_loop(self, jumps, prev_loops, loop_start):

        # The structure of a for-loop required by SpirV / our internal bytecode:
        #
        # * block zero: the block from which the loop starts
        # * header block: we only have a co_branch_loop here
        # * iter block: ending in a co_branch_conditional that goes to the body or merge block
        # * body block: the loop body
        # * continue block: may increase iter variable, jumps to header block
        # * merge block: the loop ends here

        # We only know that loop_start is the start of the "header block"
        # (in the py bytecode). So we are going to trace all the info we need ...

        # Look for jumps to the header -> to find the end, and continue's
        jumps_to_start = []
        for i, target in jumps.items():
            if target < i and target == loop_start:
                jumps_to_start.append(i)

        # Now we know the end (but there may be two positions to jump to)
        assert len(jumps_to_start) > 0
        our_ends = [jumps_to_start[-1] + 2]
        if self._peek(our_ends[0]) == "POP_BLOCK":
            our_ends.append(our_ends[0] + 2)
        ends = our_ends.copy()
        ends += [x["start"] for x in prev_loops] + [x["end"] for x in prev_loops]

        # Take a look at that first jump. If it jumps to the merge_block,
        # we have a valid iter block.
        first_jump_is_to_end = body_target = None
        for i, target in jumps.items():
            if i > loop_start:
                if target in ends:
                    first_jump_is_to_end = True
                    body_target = i + 2
                elif self._peek(target) == "BREAK_LOOP":
                    first_jump_is_to_end = True
                    body_target = i + 2
                break

        # Check what kind of loop this is
        has_for_iter = self._peek(loop_start) == "FOR_ITER"

        # Init loop info
        loop_info = {}
        loop_info["type"] = "for" if has_for_iter else "while"
        loop_info["start"] = loop_start
        loop_info["end"] = our_ends[-1]
        loop_info["first_jump_is_to_end"] = first_jump_is_to_end

        # Define the labels that we need for the loop structure
        loop_idx = len(prev_loops) + 1
        loop_info["header_label"] = f"Lh{loop_idx}"
        loop_info["iter_label"] = f"Li{loop_idx}"
        loop_info["continue_label"] = f"Lc{loop_idx}"
        loop_info["body_label"] = f"Lb{loop_idx}"
        loop_info["merge_label"] = f"Lm{loop_idx}"

        # Define label mappings
        loop_info["labelmap"] = labelmap = {}

        # The Py bytecode jumps to loop_start become branches to continue_label.
        # Also prevent continue label from being aut-created and collapsed.
        labelmap[loop_start] = loop_info["continue_label"]
        self._protected_labels.add(loop_start)

        # Any jumps to what could mean end-targets should be branches to merge_label.
        for end in ends:
            labelmap[end] = loop_info["merge_label"]
        for end in our_ends:
            self._protected_labels.add(end)

        # If we're not generating the body_label, we want it auto-emitted!
        if first_jump_is_to_end and loop_info["type"] == "while":
            self._labels[body_target] = loop_info["body_label"]
        else:
            pass  # we create a body in _start_loop()

        # Protect the custom labels for collapsing (when empty) for good measure
        for label_id in ["iter_label", "continue_label", "merge_label", "body_label"]:
            self._protected_labels.add(loop_info[label_id])

        return loop_info

    def _replace_labels(self, labels_to_replace):

        # Handle recursion
        for key in list(labels_to_replace):
            while labels_to_replace[key] in labels_to_replace:
                labels_to_replace[key] = labels_to_replace[labels_to_replace[key]]

        # Replace the labels
        for i in range(len(self._opcodes)):
            if self._opcodes[i][0] in ("co_label", "co_branch"):
                if self._opcodes[i][1] in labels_to_replace:
                    self._opcodes[i] = (
                        self._opcodes[i][0],
                        labels_to_replace[self._opcodes[i][1]],
                    )
            elif self._opcodes[i][0] in ("co_branch_conditional", "co_branch_loop"):
                op = list(self._opcodes[i])
                changed = False
                for j in range(1, len(op)):
                    if op[j] in labels_to_replace:
                        op[j] = labels_to_replace[op[j]]
                        changed = True
                if changed:
                    self._opcodes[i] = tuple(op)

    def _fix_empty_blocks(self):
        # Sometimes Python bytecode contains an empty block (i.e. code
        # jumpt to a location, from which it jumps to another location
        # immediately). In such cases, the control flow can be
        # incosistent, with some branches jumping to that empty block,
        # and some skipping it. The code below finds such empty blocks
        # and resolve them.

        labels_to_replace = {}

        def _set_new_label(label, new_label):
            while label in labels_to_replace:
                label = labels_to_replace[label]
            labels_to_replace[label] = new_label

        for i in reversed(range(len(self._opcodes) - 1)):
            if (
                self._opcodes[i][0] == "co_label"
                and self._opcodes[i + 1][0] == "co_branch"
                and self._opcodes[i][1] not in self._protected_labels
            ):
                _set_new_label(self._opcodes[i][1], self._opcodes[i + 1][1])
                self._opcodes.pop(i)
                self._opcodes.pop(i)

        self._replace_labels(labels_to_replace)

    def _fix_or_control_flow(self):
        # In `a or b` many languages don't evaluate `b` if `a` evaluates
        # to truethy. This introduces more complex control flow, with
        # multiple branches passing through the same block. SpirV does
        # not allow this. Sadly for us, the bytecode has already
        # resolved `or`'s into control flow ... so we have to detect
        # the pattern. In `a and b`, `b` is not evaluated when `a`
        # evaluates to falsy. But in this case the resulting control
        # flow is fine, and we're probably unable to detect it reliably.

        def _get_block_to_resolve():
            conditional_branches = {}
            cur_block = None
            cur_block_i = 0
            for i in range(len(self._opcodes)):
                opcode, *args = self._opcodes[i]
                if opcode == "co_label":
                    cur_block = args[0]
                    cur_block_i = i
                elif opcode == "co_branch_conditional":
                    # Detect that this conditional branch is part of an earlier comparison
                    if args[0] in conditional_branches:
                        other, ii = conditional_branches[args[0]]
                        if other == cur_block:
                            return ii, cur_block_i, i
                    elif args[1] in conditional_branches:
                        other, ii = conditional_branches[args[1]]
                        if other == cur_block:
                            return ii, cur_block_i, i
                    # Register this branch (note that this may overwrite keys, which is ok)
                    conditional_branches[args[0]] = args[1], i
                    conditional_branches[args[1]] = args[0], i

        while True:
            block = _get_block_to_resolve()
            if not block:
                break
            i_ins, i_label, i_cond = block
            # Get all the labels
            labels1 = self._opcodes[i_ins][1:]  # this label and the common block
            labels2 = self._opcodes[i_cond][1:]  # the common block and the else
            # Rip out the current label
            selection = self._opcodes[i_label + 1 : i_cond]
            self._opcodes[i_label : i_cond + 1] = []
            # Determine how to combine these
            if labels1[0] == labels2[0]:  # comp1 is true or comp2 is true
                selection.append(("co_binary_op", "or"))
                selection.append(("co_branch_conditional", labels1[0], labels2[1]))
            elif labels1[0] == labels2[1]:  # comp1 is true or comp2 is false
                selection.append(("co_unary_op", "not"))
                selection.append(("co_binary_op", "or"))
                selection.append(("co_branch_conditional", labels1[0], labels2[0]))
            elif labels1[1] == labels2[0]:  # comp1 is false or comp2 is true
                selection.insert(0, ("co_unary_op", "not"))
                selection.append(("co_binary_op", "or"))
                selection.append(("co_branch_conditional", labels1[1], labels2[1]))
            elif labels1[1] == labels2[1]:  # comp1 is false or comp2 is false
                selection.append(("co_binary_op", "and"))
                selection.append(("co_unary_op", "not"))
                selection.append(("co_branch_conditional", labels1[1], labels2[0]))
            # Put it back in with the parent label
            self._opcodes[i_ins : i_ins + 1] = selection

    def _fix_consistent_labels(self):
        # Rename the block labels, so that they are numbered in order
        # of appearance of the co_label. This also makes the resulting
        # bytecode consistent between Python versions/implementations.

        labels_to_replace = {}

        def _set_new_label(label, new_label):
            while label in labels_to_replace:
                label = labels_to_replace[label]
            labels_to_replace[label] = new_label

        count = 0
        for i in range(len(self._opcodes)):
            if self._opcodes[i][0] == "co_label":
                label = self._opcodes[i][1]
                if not label.startswith("L"):
                    count += 1
                    _set_new_label(label, f"L{count}")

        self._replace_labels(labels_to_replace)

    def _next(self):
        assert self._pointer % 2 == 0
        opcode = self._py_bytecode[self._pointer]
        arg = self._py_bytecode[self._pointer + 1]
        # Resolve name
        opcode = dis.opname[opcode]
        # Resolve EXTENDED_ARG
        n, i = 1, self._pointer
        while self._py_bytecode[i - 2] == EXTENDED_ARG:
            arg += self._py_bytecode[i - 1] * 256 ** n
            n += 1
            i -= 2
        self._pointer += 2
        return opcode, arg

    def _peek(self, pos=None):
        pos = self._pointer if pos is None else pos
        res = self._py_bytecode[pos]
        if pos % 2 == 0:
            # Resolve name
            res = dis.opname[res]
        else:
            # Resolve EXTENDED_ARG
            n, i = 1, pos - 1
            while self._py_bytecode[i - 2] == EXTENDED_ARG:
                res += self._py_bytecode[i - 1] * 256 ** n
                n += 1
                i -= 2
        return res

    def _get_label(self, pointer_pos):
        loop_labels = self._loop_stack[-1].get("labelmap", {})
        if pointer_pos in loop_labels:
            return loop_labels[pointer_pos]
        elif pointer_pos not in self._labels:
            # Labels are set to bytecode index at first. Later we turn
            # them into values that are consistent across Python
            # versions. The final label starts with "L", and labels
            # starting with "L" will not be renamed.
            self._labels[pointer_pos] = str(pointer_pos)
        return self._labels[pointer_pos]

    # %%

    def _op_extended_arg(self, arg):
        pass

    def _op_dup_top(self, arg):
        ob = self._stack_pop()
        self._stack.extend([ob, ob])
        self.emit(op.co_dup_top)

    def _op_pop_top(self, arg):
        self._stack_pop()
        self.emit(op.co_pop_top)

    def _op_rot_two(self, arg):
        ob1 = self._stack_pop()
        ob2 = self._stack_pop()
        self._stack.extend([ob1, ob2])
        self.emit(op.co_reverse_stack, 2)  # rotate and reverse are same for n = 2

    def _op_rot_three(self, arg):
        ob1 = self._stack_pop()
        ob2 = self._stack_pop()
        ob3 = self._stack_pop()
        self._stack.extend([ob1, ob3, ob2])
        self.emit(op.co_rotate_stack, 3)

    def _op_rot_four(self, arg):  # py 3.8+
        ob1 = self._stack_pop()
        ob2 = self._stack_pop()
        ob3 = self._stack_pop()
        ob4 = self._stack_pop()
        self._stack.extend([ob1, ob4, ob3, ob2])
        self.emit(op.co_rotate_stack, 4)

    def _op_return_value(self, arg):
        result = self._stack_pop()
        assert result is None
        if self._pointer == len(self._py_bytecode):
            pass
        else:
            self.emit(op.co_return)

    def _op_load_fast(self, i):
        # store a variable that is used in an inner scope.
        name = self._co.co_varnames[i]
        if name in self._input:
            self.emit(op.co_load_name, "input." + name)
            self._stack.append("input." + name)
        elif name in self._output:
            self.emit(op.co_load_name, "output." + name)
            self._stack.append("output." + name)
        elif name in self._uniform:
            self.emit(op.co_load_name, "uniform." + name)
            self._stack.append("uniform." + name)
        elif name in self._buffer:
            self.emit(op.co_load_name, "buffer." + name)
            self._stack.append("buffer." + name)
        elif name in self._sampler:
            self.emit(op.co_load_name, "sampler." + name)
            self._stack.append("sampler." + name)
        elif name in self._texture:
            self.emit(op.co_load_name, "texture." + name)
            self._stack.append("texture." + name)
        else:
            # Normal load
            self.emit(op.co_load_name, name)
            self._stack.append(name)

    def _op_store_fast(self, i):
        name = self._co.co_varnames[i]
        ob = self._stack_pop()  # noqa - ob not used
        # we don't prevent assigning to input here, that's the task of bc generator
        if name in self._input:
            self.emit(op.co_store_name, "input." + name)
        elif name in self._output:
            self.emit(op.co_store_name, "output." + name)
        elif name in self._uniform:
            self.emit(op.co_store_name, "uniform." + name)
        elif name in self._buffer:
            self.emit(op.co_store_name, "buffer." + name)
        elif name in self._sampler:
            self.emit(op.co_store_name, "sampler." + name)
        elif name in self._texture:
            self.emit(op.co_store_name, "texture." + name)
        else:
            # Normal store
            self.emit(op.co_store_name, name)

    def _op_load_const(self, i):
        ob = self._co.co_consts[i]
        if isinstance(ob, (float, int, bool)):
            self.emit(op.co_load_constant, ob)
            self._stack.append(ob)
        elif isinstance(ob, tuple):
            if self._peek() != "UNPACK_SEQUENCE":
                raise ShaderError(
                    "Const tuples are not supported (though you can do `a, b = c, d`)"
                )
            for x in ob:
                if isinstance(x, (float, int, bool)):
                    self.emit(op.co_load_constant, x)
                    self._stack.append(x)
                else:
                    raise ShaderError("Only float/int/bool constants supported.")
            self._stack.append(("tuple", len(ob)))  # signal for UNPACK_SEQUENCE
        elif ob is None:
            self._stack.append(None)  # Probably for the function return value
        else:
            raise ShaderError("Only float/int/bool constants supported.")

    def _op_load_global(self, i):
        # Loading a global in Python can mean different things. We need
        # to check here what it is, and make sure that the loaded thing
        # gets used and results in the correct emitted bytecode. We do
        # not emit code here, but we move a special value on the stack.
        # That value is a string name prepended with a dot, to indicate
        # it being global.
        #
        # When popping a value off the stack, one must indictate whether
        # globals are allowed. This happens only in call_function,
        # call_method, and load_attr. We must make sure that such
        # globals are handled correctly, and do not "slip through",
        # otherwise the user can get really strange error messages
        # because the stack is broken.

        name = self._co.co_names[i]

        if name in gpu_types_map:
            # A type definition
            self._stack.append(".type." + name)
        elif name in stdlib_func_names:
            # An stdlib function, like texture sampling, or ext instruction
            self._stack.append(".stdlib." + name)
        elif name in ("math", "stdlib"):
            # Namespaces, need load_attr on these
            self._stack.append("." + name)
        elif name in ("range",):
            # Builtin functions that we resolve in this compiler
            self._stack.append(".py." + name)
        else:
            raise ShaderError(f"Unknown variable name {name!r}")
        # todo: loading constants from the Python globals() scope
        # todo: loading other Python shader functions

    def _op_load_attr(self, i):
        name = self._co.co_names[i]
        ob = self._stack_pop(True)  # allow global
        if not isinstance(ob, str):
            # Likely vector swizzling
            self.emit(op.co_load_attr, name)
            self._stack.append(None)
        elif ob == ".stdlib":
            if name not in stdlib_func_names:
                raise ShaderError(f"No stdlib function {name}")
            self._stack.append(".stdlib." + name)  # new global on the stack
        elif ob.startswith(".math"):
            ob = getattr(math, name, None)
            if isinstance(ob, float):
                self.emit(op.co_load_constant, ob)  # e.g. math.pi
                self._stack.append(None)
            elif name == "fmod":
                self._stack.append(".py.rem")  # new global on the stack
            elif name in stdlib_func_names:
                self._stack.append(".stdlib." + name)  # new global on the stack
            else:
                raise ShaderError(f"No math constant/function {name}")
        elif ob.startswith("texture."):
            # Calling a texture sampling function as a method on a texture
            # object. Not a global! We need to communicate to call_funcion/call_method
            # that this is such a function.
            self._stack.append(ob)
            self._stack.append("texture." + name)
        elif ob.startswith("."):
            # Catch invalid use of globals
            raise ShaderError(f"Cannot load attribute '{name}' from '{ob}'")
        else:
            self.emit(op.co_load_attr, name)
            self._stack.append(None)

    def _op_load_method(self, i):
        self._stack.append(self._stack[-1])  # for _op_load_attr
        return self._op_load_attr(i)

    def _op_load_deref(self, arg):
        # ext_ob_name = self._co.co_freevars[i]
        # ext_ob = self._py_func.__closure__[i]
        raise ShaderError("Shaders cannot be used as closures atm.")

    def _op_store_attr(self, i):
        name = self._co.co_names[i]
        ob = self._stack_pop()
        value = self._stack_pop()  # noqa
        raise ShaderError(f"{ob}.{name} store")

    def _op_call_function(self, nargs):
        args = [self._stack_pop(True) for i in range(nargs)]
        args.reverse()

        func = self._stack_pop(True)  # allow global

        if not isinstance(func, str):
            raise ShaderError(f"Cannot call object '{func}'.")
        self._call_function(func, args)

    def _op_call_method(self, nargs):
        args = [self._stack_pop() for i in range(nargs)]
        args.reverse()

        func = self._stack_pop(True)  # allow global

        ob = self._stack_pop(True)  # noqa - need to get rid of this here
        assert func.startswith("texture.") or func.startswith(".")

        assert isinstance(func, str)
        self._call_function(func, args)

    def _call_function(self, func, args):
        nargs = len(args)
        funcname = func.split(".", 2)[-1]

        # Args can be globals, but only for e.g. Vector(2, f32)
        if func.startswith(".type."):
            args, ori_args = [], args
            for arg in ori_args:
                if isinstance(arg, str) and arg.startswith(".type."):
                    args.append(arg[6:])
                else:
                    args.append(arg)
        for arg in args:
            if isinstance(arg, str) and arg.startswith("."):
                raise ShaderError(f"Cannot call {func} with arg {arg}")

        if func.startswith("texture."):
            # A texture function called as a method of a texture object
            # This is syntactic sugar. We just need to increase nargs.
            ob = self._stack_pop()
            assert ob.startswith("texture.")  # a texture object
            self.emit(op.co_call, funcname, nargs + 1)
            self._stack.append(None)
        elif func.startswith(".type."):
            # A type definition
            if "(" not in funcname and gpu_types_map[funcname].is_abstract:
                type_str = f"{funcname}({','.join(str(arg) for arg in args)})"
                self._stack.append(".type." + type_str)
            else:
                self.emit(op.co_call, funcname, nargs)
                self._stack.append(None)
        elif func == ".py.rem":
            assert nargs == 2
            self.emit(op.co_binary_op, "rem")
            self._stack.append(None)
        elif func == ".py.range":
            if not (
                self._peek(self._pointer) == "GET_ITER"
                and self._peek(self._pointer + 2) == "FOR_ITER"
            ):
                raise ShaderError("range() can only be used as a for-loop iter.")
            loop_info = self._loops_to_handle[0]
            assert loop_info["start"] == self._pointer + 2
            loop_info["range_is_set"] = True
            if nargs == 1:
                self.emit(op.co_load_constant, 0)
                self.emit(op.co_reverse_stack, 2)
                self.emit(op.co_load_constant, 1)
            elif nargs == 2:
                self.emit(op.co_load_constant, 1)
            elif nargs == 3:
                step = args[2]
                if not (isinstance(step, int) and step > 0):
                    raise ShaderError("range() step must be a constant int > 0")
            else:
                raise ShaderError("range() must have 1, 2 or 3 args.")
            self._stack.append("range")
            # nothing to emit yet
        elif func.startswith((".stdlib.", ".math.")):
            self.emit(op.co_call, funcname, nargs)
            self._stack.append(None)
        elif func.startswith("."):
            raise ShaderError(f"Unknown external function {func}.")
        else:
            raise ShaderError(f"Cannot call object {func}.")

    def _op_binary_subscr(self, arg):
        index = self._stack_pop()
        ob = self._stack_pop()  # noqa - ob not ised
        if isinstance(index, tuple):
            self.emit(op.co_load_index, len(index))
        else:
            self.emit(op.co_load_index)
        self._stack.append(None)

    def _op_store_subscr(self, arg):
        index = self._stack_pop()  # noqa
        ob = self._stack_pop()  # noqa
        val = self._stack_pop()  # noqa
        self.emit(op.co_store_index)

    def _op_build_tuple(self, n):
        if self._peek() == "UNPACK_SEQUENCE":
            # We don't actually build a tuple, but mark that the stack has the values
            self._stack.append(("tuple", n))
        else:
            raise ShaderError(
                "Tuples are not supported (though you can do `a, b = c, d`)"
            )

    def _op_unpack_sequence(self, n):
        x = self._stack_pop()
        if isinstance(x, tuple) and x and x[0] == "tuple":
            # If the number of elements matches, we are all good
            if x[1] == n:
                self.emit(op.co_reverse_stack, n)
                objects = [self._stack.pop() for i in range(n)]
                self._stack.extend(objects)
            else:
                raise ShaderError(f"Cannot unpack a {x[1]} tuple into a {n}-tuple")
        else:
            raise ShaderError(
                "Cannot unpack arbitrary sequences (though you can do `a, b = c, d`)"
            )

    def _op_build_list(self, n):
        # Litaral list
        res = [self._stack_pop() for i in range(n)]
        res = list(reversed(res))
        self._stack.append(res)
        self.emit(op.co_load_array, n)

    def _op_build_map(self, arg):
        raise ShaderError("Dict not allowed in Shader-Python")

    def _op_build_const_key_map(self, arg):
        # The version of BUILD_MAP specialized for constant keys. Py3.6+
        raise ShaderError("Dict not allowed in Shader-Python")

    def _op_unary_positive(self, arg):
        self._stack_pop()
        self._stack.append(None)
        # this is a no-op

    def _op_unary_negative(self, arg):
        self._stack_pop()
        self._stack.append(None)
        self.emit(op.co_unary_op, "neg")

    def _op_unary_not(self, arg):
        self._stack_pop()
        self._stack.append(None)
        self.emit(op.co_unary_op, "not")

    def _binary_op(self, binop):
        self._stack_pop()
        self._stack_pop()
        self._stack.append(None)
        self.emit(op.co_binary_op, binop)

    def _inplace_op(self, binop):
        val = self._stack_pop()  # noqa
        name = self._stack_pop()
        self._stack.append(None)
        assert isinstance(name, str)
        self.emit(op.co_binary_op, binop)

    def _op_inplace_add(self, arg):
        self._inplace_op("add")

    def _op_inplace_subtract(self, arg):
        self._inplace_op("sub")

    def _op_inplace_multiply(self, arg):
        self._inplace_op("mul")

    def _op_inplace_true_divide(self, arg):
        self._inplace_op("fdiv")

    def _op_inplace_floor_divide(self, arg):
        self._inplace_op("idiv")

    def _op_binary_add(self, arg):
        self._binary_op("add")

    def _op_binary_subtract(self, arg):
        self._binary_op("sub")

    def _op_binary_multiply(self, arg):
        self._binary_op("mul")

    def _op_binary_matrix_multiply(self, arg):
        self._binary_op("mmul")

    def _op_binary_true_divide(self, arg):
        # We use the fdiv opcode that only works for floats. Python
        # auto-converts ints to float when dividing. A shader does not.
        # To avoid confusion, users have to use the normal division for
        # floats, and the // division for ints.
        self._binary_op("fdiv")

    def _op_binary_floor_divide(self, arg):
        self._binary_op("idiv")

    def _op_binary_power(self, arg):
        exp = self._stack_pop()
        self._stack_pop()  # base
        self._stack.append(None)
        if exp == 2:  # shortcut
            self.emit(op.co_pop_top)
            self.emit(op.co_dup_top)
            self.emit(op.co_binary_op, "mul")
        elif exp == 0.5:
            self.emit(op.co_pop_top)
            self.emit(op.co_call, "sqrt", 1)
        else:
            self.emit(op.co_call, "pow", 2)

    def _op_binary_modulo(self, arg):
        self._binary_op("mod")

    def _op_compare_op(self, arg):
        cmp = cmp_op[arg]
        if cmp not in ("<", "<=", "==", "!=", ">", ">="):
            raise ShaderError(f"Compare op {cmp} not supported in shaders.")
        self._stack_pop()
        self._stack_pop()
        self._stack.append(None)
        self.emit(op.co_compare, cmp)

    def _op_jump_absolute(self, target):
        label = self._get_label(target)
        if label.startswith("Lm") and self._opcodes[-1][0] == "co_pop_top":
            # This is a break in Python 3.8+ - I think it pops the iterator
            self._opcodes.pop(-1)
        self.emit(op.co_branch, label)

    def _op_jump_forward(self, delta):
        target = self._pointer + delta
        if self._opcodes[-1][0].startswith("co_branch"):
            # Is this a Python bug? Below is a snippet of seen Python bytecode.
            # There are no jumps to 28. Maybe there *could* be? If so, we would
            # emit a co_label, and this IF wouldn't triger (and all is well).
            # 26 JUMP_ABSOLUTE           14
            # 28 JUMP_FORWARD            10 (to 40)
            return
        self.emit(op.co_branch, self._get_label(target))

    def _op_pop_jump_if_false(self, target):
        condition = self._stack_pop()  # noqa
        self.emit(
            op.co_branch_conditional,
            self._get_label(self._pointer),
            self._get_label(target),
        )
        # todo: spirv supports hints on what branch is the most likely

    def _op_pop_jump_if_true(self, target):
        condition = self._stack_pop()  # noqa
        self.emit(
            op.co_branch_conditional,
            self._get_label(target),
            self._get_label(self._pointer),
        )

    def _op_jump_if_true_or_pop(self, target):
        # This is xx OR yy, but only when a result is needed
        # So not inside ``if xx or yy:``, but in ``if bool(xx or yy):``

        # The xx is now on the stack. In the next instructions yy will be
        # pushed on the stack, and at target, we continue. That's where we
        # need to insert the OR.

        # self._insert_at[target] = ("co_binary_op", "or")

        # ... except that determining if an arbitrary object is true
        # or false is not trivial. We could add something like co_bool,
        # but maybe we should avoid that temptation, as it does not fit
        # a strongly typed language well ...
        raise ShaderError(
            "Implicit bool conversions not supported. Maybe use ``x if y else z``?"
        )

    def _op_jump_if_false_or_pop(self, target):
        # Same as _op_jump_if_true_or_pop, but for AND
        # self._insert_at[target] = ("co_binary_op", "and")
        raise ShaderError(
            "Implicit bool conversions not supported. Maybe use ``x if y else z``?"
        )

    def _start_loop(self, loop_info):

        # This gets called right before the first instruction of the loop
        # gets processed. We need to emit some loop-related code here.

        self._loop_stack.append(loop_info)

        if loop_info["type"] == "for":
            # Check that the range is set
            if not loop_info.get("range_is_set"):
                raise ShaderError("Shader for-loop must use range()")

            # Consume next codepoint - the storing of the iter value
            assert self._peek(self._pointer) == "FOR_ITER"
            assert self._peek(self._pointer + 2) == "STORE_FAST"
            iter_name_index = self._peek(self._pointer + 2 + 1)
            iter_name = self._co.co_varnames[iter_name_index]
            loop_info["iter_name"] = iter_name

            # Block 0 (the current block) - prepare iter variable
            # Note that in the range() call, we've put three variables on the stack
            self.emit(op.co_store_name, iter_name + "-step")
            self.emit(op.co_store_name, iter_name + "-stop")
            self.emit(op.co_store_name, iter_name + "-start")
            self.emit(op.co_load_name, iter_name + "-start")
            self.emit(op.co_store_name, iter_name)
            self.emit(op.co_branch, loop_info["header_label"])
            # Block 1 - the "header" of the loop
            self.emit(op.co_label, loop_info["header_label"])
            self.emit(
                op.co_branch_loop,
                loop_info["iter_label"],
                loop_info["continue_label"],
                loop_info["merge_label"],
            )
            # Block 2 - the block that decides whether to break from the loop
            self.emit(op.co_label, loop_info["iter_label"])
            self.emit(op.co_load_name, iter_name)
            self.emit(op.co_load_name, iter_name + "-stop")
            self.emit(op.co_compare, "<")
            self.emit(
                op.co_branch_conditional,
                loop_info["body_label"],
                loop_info["merge_label"],
            )
            # Block 3 - the body (can consist of more blocks)
            self.emit(op.co_label, loop_info["body_label"])
            # ... the body is what gets processed next
            # The continue_label and merge_label get emitted in _end_loop

        elif loop_info["type"] == "while":

            # Block 0 - the current block
            self.emit(op.co_branch, loop_info["header_label"])
            # Block 1 - the "header" of the loop
            self.emit(op.co_label, loop_info["header_label"])
            self.emit(
                op.co_branch_loop,
                loop_info["iter_label"],
                loop_info["continue_label"],
                loop_info["merge_label"],
            )
            # Block 2 - the block that decides whether to break from the loop
            self.emit(op.co_label, loop_info["iter_label"])
            if loop_info["first_jump_is_to_end"]:
                # The self._labels[target] = loop_info["body_label"] has been applied,
                # so the body label (and the branch to it) get generated as we go.
                pass
            else:
                self.emit(op.co_load_constant, True)
                self.emit(
                    op.co_branch_conditional,
                    loop_info["body_label"],
                    loop_info["merge_label"],
                )
                self.emit(op.co_label, loop_info["body_label"])
            # The continue_label and merge_label get emitted in _end_loop

        else:
            raise RuntimeError(f"invalid loop type {loop_info['type'] }")

    def _end_loop(self):

        # This gets called right before the first instruction after the loop.
        # We need to emit some instructions to close up the loop.

        loop_info = self._loop_stack.pop(-1)

        if loop_info["type"] == "for":
            # For-loop: this is where the iter value is incremented.
            iter_name = loop_info["iter_name"]
            self.emit(op.co_label, loop_info["continue_label"])
            self.emit(op.co_load_name, iter_name)
            self.emit(op.co_load_name, iter_name + "-step")
            self.emit(op.co_binary_op, "add")
            self.emit(op.co_store_name, iter_name)
            self.emit(op.co_branch, loop_info["header_label"])
            self.emit(op.co_label, loop_info["merge_label"])
        else:
            # While-loop: just jump to the header. We add two no-op instruction
            # to avoid the branch from being collapsed by our fix_empty_blocks()
            # Note that this does not cause any SPIRV code (except
            # perhaps an unused definition of a constant 0.0)
            self.emit(op.co_label, loop_info["continue_label"])
            self.emit(op.co_branch, loop_info["header_label"])
            self.emit(op.co_label, loop_info["merge_label"])

    def _op_setup_loop(self, delta):
        # This is Python < 3.8 indicating that there is a loop coming. We don't use it.
        self._pointer + delta
        assert self._loops_to_handle[0]["end"]

    def _op_break_loop(self, arg):
        # Python < 3.8
        self.emit(op.co_branch, self._loop_stack[-1]["merge_label"])

    def _op_continue_loop(self, target):
        # This bytecode op is present in Python < 3.8, but does not seem to be
        # used in 3.6 and 3.7 either ...
        target1 = target  # for-iter
        target2 = self._loop_stack[-1]["continue_label"]
        assert target1 == target2
        self.emit(op.co_branch, target2)

    def _op_get_iter(self, arg):
        func = self._stack_pop()
        if func != "range":
            raise ShaderError("Can only use a loop with range()")
        self._stack.append(func)
        # Note: in op_call_function we've already made sure that there are three arg values on the stack

    def _op_for_iter(self, delta):
        # This is the start of a for-loop, but we don't trigger using the method,
        # because our logic needs to take while-loops into account too.
        # But we can do some checks for good measure :)

        target = self._pointer + delta
        here = self._pointer - 2

        next_op, next_val = self._next()  # STORE_FAST, iter variable name
        loop_info = self._loop_stack[-1]

        assert here == loop_info["start"]
        assert target in (loop_info["end"], loop_info["end"] - 2)
        assert next_op == "STORE_FAST"
        assert self._co.co_varnames[next_val] == loop_info["iter_name"]

    def _op_pop_block(self, arg):
        pass



"""
Python opcodes

Taken from https://github.com/beeware/batavia/blob/master/batavia/modules/dis.js
which is Copyright Russell Keith-Magee, and licensed under the 3-clause BSD.

Added some Pypy opcodes too.
"""

# NOTE: THIS CODE IS AUTOGENERATED; DO NOT EDIT


class dis:

    hasconst = {}
    hasname = {}
    hasjrel = {}
    hasjabs = {}
    haslocal = {}
    hascompare = {}
    hasfree = {}
    hasnargs = {}

    opmap = {}
    opname = []

    unary_ops = {}
    binary_ops = {}
    inplace_ops = {}
    # slice_ops: {},


def def_op(name, op):
    dis.opname[op] = name
    dis.opmap[name] = op


def def_unary_op(name, op):
    def_op(name, op)
    dis.unary_ops[op] = op


def def_binary_op(name, op):
    def_op(name, op)
    dis.binary_ops[op] = op


def def_inplace_op(name, op):
    def_op(name, op)
    dis.inplace_ops[op] = op


# function def_slice_op(name, op) {
#     def_op(name, op);
#     slice_ops[op] = op;
# }


def name_op(name, op):
    def_op(name, op)
    dis.hasname[op] = op


def jrel_op(name, op):
    def_op(name, op)
    dis.hasjrel[op] = op


def jabs_op(name, op):
    def_op(name, op)
    dis.hasjabs[op] = op


# Prime the opname list with all possible opnames
for op in range(256):
    dis.opname.append("<" + str(op) + ">")

# Register the known opnames
def_op("POP_TOP", 1)
def_op("ROT_TWO", 2)
def_op("ROT_THREE", 3)
def_op("DUP_TOP", 4)
def_op("DUP_TOP_TWO", 5)

# Introduced in Python 3.8
def_op("ROT_FOUR", 6)

def_op("NOP", 9)
dis.NOP = 9  # TODO why does this require special handling?
def_unary_op("UNARY_POSITIVE", 10)
def_unary_op("UNARY_NEGATIVE", 11)
def_unary_op("UNARY_NOT", 12)

def_unary_op("UNARY_INVERT", 15)

# Introduced in Python 3.5
def_binary_op("BINARY_MATRIX_MULTIPLY", 16)
def_inplace_op("INPLACE_MATRIX_MULTIPLY", 17)

def_binary_op("BINARY_POWER", 19)
def_binary_op("BINARY_MULTIPLY", 20)

def_binary_op("BINARY_MODULO", 22)
def_binary_op("BINARY_ADD", 23)
def_binary_op("BINARY_SUBTRACT", 24)
def_binary_op("BINARY_SUBSCR", 25)
def_binary_op("BINARY_FLOOR_DIVIDE", 26)
def_binary_op("BINARY_TRUE_DIVIDE", 27)
def_inplace_op("INPLACE_FLOOR_DIVIDE", 28)
def_inplace_op("INPLACE_TRUE_DIVIDE", 29)

# Introduced in Python 3.5
def_op("GET_AITER", 50)
def_op("GET_ANEXT", 51)
def_op("BEFORE_ASYNC_WITH", 52)

# Introduced in Python 3.8
def_op("BEGIN_FINALLY", 53)
def_op("END_ASYNC_FOR", 54)

def_inplace_op("INPLACE_ADD", 55)
def_inplace_op("INPLACE_SUBTRACT", 56)
def_inplace_op("INPLACE_MULTIPLY", 57)

def_inplace_op("INPLACE_MODULO", 59)
def_op("STORE_SUBSCR", 60)
def_op("DELETE_SUBSCR", 61)
def_binary_op("BINARY_LSHIFT", 62)
def_binary_op("BINARY_RSHIFT", 63)
def_binary_op("BINARY_AND", 64)
def_binary_op("BINARY_XOR", 65)
def_binary_op("BINARY_OR", 66)
def_inplace_op("INPLACE_POWER", 67)
def_op("GET_ITER", 68)

# Introduced in Python 3.5
def_op("GET_YIELD_FROM_ITER", 69)

def_op("PRINT_EXPR", 70)
def_op("LOAD_BUILD_CLASS", 71)
def_op("YIELD_FROM", 72)

# Introduced in Python 3.5
def_op("GET_AWAITABLE", 73)

def_inplace_op("INPLACE_LSHIFT", 75)
def_inplace_op("INPLACE_RSHIFT", 76)
def_inplace_op("INPLACE_AND", 77)
def_inplace_op("INPLACE_XOR", 78)
def_inplace_op("INPLACE_OR", 79)
def_op("BREAK_LOOP", 80)
def_op("WITH_CLEANUP", 81)

# Introduced in Python 3.5
def_op("WITH_CLEANUP_FINISH", 82)

def_op("RETURN_VALUE", 83)
def_op("IMPORT_STAR", 84)

# Introduced in Python 3.6
def_op("SETUP_ANNOTATIONS", 85)

def_op("YIELD_VALUE", 86)
def_op("POP_BLOCK", 87)
def_op("END_FINALLY", 88)
def_op("POP_EXCEPT", 89)

dis.HAVE_ARGUMENT = 90  # Opcodes from here have an argument:

name_op("STORE_NAME", 90)  # Index in name list
name_op("DELETE_NAME", 91)  # ""
def_op("UNPACK_SEQUENCE", 92)  # Number of tuple items
jrel_op("FOR_ITER", 93)
def_op("UNPACK_EX", 94)
name_op("STORE_ATTR", 95)  # Index in name list
name_op("DELETE_ATTR", 96)  # ""
name_op("STORE_GLOBAL", 97)  # ""
name_op("DELETE_GLOBAL", 98)  # ""
def_op("LOAD_CONST", 100)  # Index in const list
dis.hasconst[100] = 100
name_op("LOAD_NAME", 101)  # Index in name list
def_op("BUILD_TUPLE", 102)  # Number of tuple items
def_op("BUILD_LIST", 103)  # Number of list items
def_op("BUILD_SET", 104)  # Number of set items
def_op("BUILD_MAP", 105)  # Number of dict entries (upto 255)
name_op("LOAD_ATTR", 106)  # Index in name list
def_op("COMPARE_OP", 107)  # Comparison operator
dis.hascompare[107] = 107
name_op("IMPORT_NAME", 108)  # Index in name list
name_op("IMPORT_FROM", 109)  # Index in name list

jrel_op("JUMP_FORWARD", 110)  # Number of bytes to skip
jabs_op("JUMP_IF_FALSE_OR_POP", 111)  # Target byte offset from beginning of code
jabs_op("JUMP_IF_TRUE_OR_POP", 112)  # ""
jabs_op("JUMP_ABSOLUTE", 113)  # ""
jabs_op("POP_JUMP_IF_FALSE", 114)  # ""
jabs_op("POP_JUMP_IF_TRUE", 115)  # ""

name_op("LOAD_GLOBAL", 116)  # Index in name list

jabs_op("CONTINUE_LOOP", 119)  # Target address
jrel_op("SETUP_LOOP", 120)  # Distance to target address
jrel_op("SETUP_EXCEPT", 121)  # ""
jrel_op("SETUP_FINALLY", 122)  # ""

def_op("LOAD_FAST", 124)  # Local variable number
dis.haslocal[124] = 124
def_op("STORE_FAST", 125)  # Local variable number
dis.haslocal[125] = 125
def_op("DELETE_FAST", 126)  # Local variable number
dis.haslocal[126] = 126

def_op("RAISE_VARARGS", 130)  # Number of raise arguments (1, 2, or 3);
def_op("CALL_FUNCTION", 131)  # #args + (#kwargs << 8);
dis.hasnargs[131] = 131
def_op("MAKE_FUNCTION", 132)  # Number of args with default values
def_op("BUILD_SLICE", 133)  # Number of items
def_op("MAKE_CLOSURE", 134)
def_op("LOAD_CLOSURE", 135)
dis.hasfree[135] = 135
def_op("LOAD_DEREF", 136)
dis.hasfree[136] = 136
def_op("STORE_DEREF", 137)
dis.hasfree[137] = 137
def_op("DELETE_DEREF", 138)
dis.hasfree[138] = 138

def_op("CALL_FUNCTION_VAR", 140)  # #args + (#kwargs << 8);
dis.hasnargs[140] = 140
def_op("CALL_FUNCTION_KW", 141)  # #args + (#kwargs << 8);
dis.hasnargs[141] = 141
def_op("CALL_FUNCTION_VAR_KW", 142)  # #args + (#kwargs << 8);
dis.hasnargs[142] = 142

jrel_op("SETUP_WITH", 143)
def_op("EXTENDED_ARG", 144)
dis.EXTENDED_ARG = 144
def_op("LIST_APPEND", 145)
def_op("SET_ADD", 146)
def_op("MAP_ADD", 147)

def_op("LOAD_CLASSDEREF", 148)
dis.hasfree[148] = 148

# Introduced in Python 3.5
def_op("BUILD_LIST_UNPACK", 149)
def_op("BUILD_MAP_UNPACK", 150)
def_op("BUILD_MAP_UNPACK_WITH_CALL", 151)
def_op("BUILD_TUPLE_UNPACK", 152)
def_op("BUILD_SET_UNPACK", 153)
def_op("SETUP_ASYNC_WITH", 154)
def_op("FORMAT_VALUE", 155)

# Introduced in Python 3.6
def_op("BUILD_CONST_KEY_MAP", 156)

# Introduced in Python 3.7
def_op("BUILD_STRING", 157)
def_op("BUILD_TUPLE_UNPACK_WITH_CALL", 158)
name_op("LOAD_METHOD", 160)
def_op("CALL_METHOD", 161)
dis.hasnargs[161] = 161

# Pypy: 3.6.9 (PyPy 7.3.0)
name_op("LOAD_METHOD", 201)
def_op("CALL_METHOD", 202)

# Introduced in Python 3.8
def_op("CALL_FINALLY", 162)
def_op("POP_FINALLY", 163)



"""
Types on the GPU (from SpirV spec):

* Basic types are boolean, int, float. Latter two are numerics, all three are scalars.
* Vector is two or more values of scalars (float, int, bool). For lengt > 4 need capabilities.
* Matrix is 2, 3, or 4 float vectors (each vector is a column).
* Array is homogeneous collection of non-void-type objects.
* Structure is heterogeneous collection of non-void-type objects.
* image, sampler, ...

Here we follow the SpirV type hierarchy. We define abstract types, which
can be specialized (made concrete) by calling them. By convention,
abstract types start with a capital letter, concrete types are lowercase.

"""

import ctypes


_subtypes = {}


def _create_type(name, base, props):
    """ Create a new type, memoize on name.
    """
    if name not in _subtypes:
        assert not props.get("is_abstract", True), "can only create concrete types"
        _subtypes[name] = type(name, (base,), props)
    return _subtypes[name]


def type_from_name(name):
    """ Get a ShaderType from its name.
    """
    original_name = name
    name = name.replace(" ", "").lower()
    return _type_from_name(name, original_name)


def _type_from_name(name, original_name):
    if name in _subtypes:
        return _subtypes[name]
    elif name.startswith("vector"):
        inner, commas = _select_between_braces(name[6:], original_name)
        assert len(commas) == 1
        n, _, subtypestr = inner.partition(",")
        subtype = _type_from_name(subtypestr, original_name)
        return Vector(int(n), subtype)
    elif name.startswith("matrix"):
        inner, commas = _select_between_braces(name[6:], original_name)
        assert len(commas) == 2
        cols, rows, subtypestr = inner.split(",", 2)
        subtype = _type_from_name(subtypestr, original_name)
        return Matrix(int(cols), int(rows), subtype)
    elif name.startswith("array"):
        inner, commas = _select_between_braces(name[5:], original_name)
        assert len(commas) in (0, 1)
        if len(commas) == 0:  # subtypestr = inner
            subtype = _type_from_name(inner, original_name)
            return Array(subtype)
        else:
            n, _, subtypestr = inner.partition(",")
            subtype = _type_from_name(subtypestr, original_name)
            return Array(int(n), subtype)
    elif name.startswith("struct"):
        inner, commas = _select_between_braces(name[6:], original_name)
        commas.insert(0, -1)
        commas.append(999999)
        parts = [inner[commas[i] + 1 : commas[i + 1]] for i in range(len(commas) - 1)]
        fields = {}
        for part in [part for part in parts if part]:
            key, _, subtypestr = part.partition("=")
            fields[key.strip()] = _type_from_name(subtypestr, original_name)
        return Struct(**fields)
    else:
        raise TypeError(f"Invalid ShaderType string '{original_name}': '{name}'")


def _select_between_braces(s, original_name):
    """ Assuming s starts with an opening brace, return the part between
    braces and the position of the comma's at the root level.
    """
    assert s[0] == "("
    level = 0
    commas = []
    for i in range(len(s)):
        if s[i] == "(":
            level += 1
        elif s[i] == ")":
            level -= 1
            if level == 0:
                break
        elif level == 1 and s[i] == ",":
            commas.append(i - 1)  # 1 offset because we drop the first char
    if level != 0:
        raise TypeError(f"No end-brace in ShaderType string '{original_name}': '{s}'")
    return s[1:i], commas


def shadertype_as_ctype(shadertype):
    """ Get a ctypes type equivalent to the given ShaderType.
    """
    if isinstance(shadertype, str):
        shadertype = type_from_name(shadertype)
    if not isinstance(shadertype, type) and issubclass(shadertype, ShaderType):
        raise TypeError("Expected str or ShaderType subclass.")
    if hasattr(shadertype, "_as_ctype"):
        return shadertype._as_ctype()
    else:
        return shadertype._ctype


# %% Really abstract types


class ShaderType:
    """ The root base class of all GPU types.
    """

    is_abstract = True

    def __init__(self):
        if self.is_abstract:
            name = self.__class__.__name__
            raise RuntimeError(
                f"Cannot instantiate {name} because it is an abstract class."
            )
        else:
            name = self.__class__.__name__
            raise RuntimeError(f"Cannot instantiate ShaderType subclass {name} (yet).")

    @classmethod
    def _as_ctype(cls):
        return cls._ctype


class Scalar(ShaderType):
    """ Base class for scalar types (float, int, bool).
    """


class Numeric(Scalar):
    """ Base class for numeric scalars (float, int).
    """


class Float(Numeric):
    """ Base class for float numerics (f16, f32, f64).
    """


class Int(Numeric):
    """ Base class for int numerics (i16, i32, i64).
    """


class Composite(ShaderType):
    """ Base class for composite types (Vector, Matrix, Aggregates).
    """


class Aggregate(Composite):
    """ Base class for Array and Struct types.
    """


# %% Abstract types (but can be used to construct composite types)


class Vector(Composite):
    """ Base class for Vector types. Concrete types are templated based on
    length and subtype.
    """

    subtype = None
    length = 0

    def __new__(cls, *args):
        if cls.is_abstract:
            if len(args) != 2:
                raise TypeError(
                    "Vector specialization needs 2 args: Vector(n, subtype)"
                )
            n, subtype = args
            n = int(n)
            if not isinstance(subtype, type) and issubclass(subtype, Scalar):
                raise TypeError("Vector subtype must be a Scalar type.")
            elif subtype.is_abstract:
                raise TypeError("Vector subtype cannot be an abstract ShaderType.")
            if n < 2 or n > 4:
                raise TypeError("Vector can have 2, 3 or 4 elements.")
            props = dict(subtype=subtype, length=n, is_abstract=False)
            return _create_type(f"Vector({n},{subtype.__name__})", Vector, props)
        else:
            return super().__new__(cls, *args)

    @classmethod
    def _as_ctype(cls):
        return cls.subtype._ctype * cls.length


class Matrix(Composite):
    """ Base class for Matrix types. Concrete types are templated based on
    cols, rows and subtype. Subtype can only be Float.
    """

    subtype = None
    cols = 0
    rows = 0

    def __new__(cls, *args):
        if cls.is_abstract:
            if len(args) != 3:
                raise TypeError(
                    "Matrix specialization needs 3 args: Matrix(cols, rows, subtype)"
                )
            cols, rows, subtype = args
            cols, rows = int(cols), int(rows)
            if not isinstance(subtype, type) and issubclass(subtype, Float):
                raise TypeError("Matrix subtype must be a Float type.")
            elif subtype.is_abstract:
                raise TypeError("Matrix subtype cannot be an abstract ShaderType.")
            if cols < 2 or cols > 4:
                raise TypeError("Matrix can have 2, 3 or 4 columns.")
            if rows < 2 or rows > 4:
                raise TypeError("Matrix can have 2, 3 or 4 rows.")
            props = dict(subtype=subtype, cols=cols, rows=rows, is_abstract=False)
            return _create_type(
                f"Matrix({cols},{rows},{subtype.__name__})", Matrix, props
            )
        else:
            return super().__new__(cls, *args)

    @classmethod
    def _as_ctype(cls):
        return cls.subtype._ctype * (cls.cols * cls.rows)  # C-contiguous


class Array(Aggregate):
    """ Base class for Array types. Concrete types are templated based on
    length and subtype. Subtype can be any ShaderType except void.
    """

    subtype = None
    length = 0

    def __new__(cls, *args):
        if cls.is_abstract:
            if len(args) == 1:
                n = 0
                subtype = args[0]
            elif len(args) == 2:
                n, subtype = args
                n = int(n)
                if n < 1:
                    raise TypeError("Array must have at least 1 element.")
            else:
                raise TypeError("Array specialization needs 2 args: Array(n, subtype)")
            # Validate
            if not isinstance(subtype, type) and issubclass(subtype, ShaderType):
                raise TypeError("Array subtype must be a ShaderType.")
            elif issubclass(subtype, void):
                raise TypeError("Array subtype cannot be void.")
            elif subtype.is_abstract:
                raise TypeError("Array subtype cannot be an abstract ShaderType.")
            props = dict(subtype=subtype, length=n, is_abstract=False)
            if n == 0:  # means it's length is unknown)
                return _create_type(f"Array({subtype.__name__})", Array, props)
            else:
                return _create_type(f"Array({n},{subtype.__name__})", Array, props)
        else:
            return super().__new__(cls, *args)

    @classmethod
    def _as_ctype(cls):
        sub_ctype = cls.subtype._as_ctype()
        return sub_ctype * cls.length


class Struct(Aggregate):
    """ Base class for Struct types. Not implemented.
    """

    def __new__(cls, **kwargs):
        if cls.is_abstract:
            n = len(kwargs)
            # Validate
            for key, subtype in kwargs.items():
                if not isinstance(subtype, type) and issubclass(subtype, ShaderType):
                    raise TypeError("Struct subtype must be a ShaderType.")
                elif issubclass(subtype, void):
                    raise TypeError("Struct subtype cannot be void.")
                elif subtype.is_abstract:
                    raise TypeError("Struct subtype cannot be an abstract ShaderType.")
                if not isinstance(
                    key, str
                ):  # and key.isidentifier(): -> allow . in name?
                    raise TypeError("Struct keys must be str.")
            # Return type
            keys = tuple(kwargs.keys())
            type_names = [
                f"{key}={subtype.__name__}" for key, subtype in kwargs.items()
            ]
            props = kwargs.copy()
            props.update(dict(length=n, keys=keys, _kwargs=kwargs, is_abstract=False))
            return _create_type(f"Struct({','.join(type_names)})", Struct, props)
        else:
            return super().__new__(cls, **kwargs)

    @classmethod
    def _as_ctype(cls):
        type_fields = [(key, val._as_ctype()) for key, val in cls._kwargs.items()]
        type_name = "C_" + cls.__name__
        return type(type_name, (ctypes.Structure,), {"_fields_": type_fields})

    @classmethod
    def get_subtype(cls, key):
        if isinstance(key, int):
            return cls._kwargs[cls.keys[key]]
        else:
            return cls._kwargs[key]


# The base types that can be used to create composite types
base_types = dict(Vector=Vector, Matrix=Matrix, Array=Array, Struct=Struct)


# %% Concrete leaf types


class void(ShaderType):
    is_abstract = False
    _ctype = ctypes.c_void_p


class boolean(Scalar):
    is_abstract = False
    _ctype = ctypes.c_bool


class f16(Float):
    is_abstract = False
    # _ctype = ctypes.c_float16 ??  maybe use uin16 and map f16 onto that data?


class f32(Float):
    is_abstract = False
    _ctype = ctypes.c_float


class f64(Float):
    is_abstract = False
    _ctype = ctypes.c_double


# For now, we simply have 3 signed ints,
# and an unsigned byte for when things need to be compact.


class u8(Int):
    is_abstract = False
    _ctype = ctypes.c_uint8


class i16(Int):
    is_abstract = False
    _ctype = ctypes.c_int16


class i32(Int):
    is_abstract = False
    _ctype = ctypes.c_int32


class i64(Int):
    is_abstract = False
    _ctype = ctypes.c_int64


# Types that are at the leaf of a composite type
leaf_types = dict(
    void=void,
    boolean=boolean,
    u8=u8,
    i16=i16,
    i32=i32,
    i64=i64,
    f16=f16,
    f32=f32,
    f64=f64,
)
_subtypes.update(leaf_types)


# %% Convenient concrete types

vec2 = Vector(2, f32)
vec3 = Vector(3, f32)
vec4 = Vector(4, f32)

ivec2 = Vector(2, i32)
ivec3 = Vector(3, i32)
ivec4 = Vector(4, i32)

bvec2 = Vector(2, boolean)
bvec3 = Vector(3, boolean)
bvec4 = Vector(4, boolean)

mat2 = Matrix(2, 2, f32)
mat3 = Matrix(3, 3, f32)
mat4 = Matrix(4, 4, f32)

mat2x2 = Matrix(2, 2, f32)
mat3x2 = Matrix(3, 2, f32)
mat4x2 = Matrix(4, 2, f32)
mat2x3 = Matrix(2, 3, f32)
mat3x3 = Matrix(3, 3, f32)
mat4x3 = Matrix(4, 3, f32)
mat2x4 = Matrix(2, 4, f32)
mat3x4 = Matrix(3, 4, f32)
mat4x4 = Matrix(4, 4, f32)

convenience_types = dict(
    vec2=vec2,
    vec3=vec3,
    vec4=vec4,
    ivec2=ivec2,
    ivec3=ivec3,
    ivec4=ivec4,
    bvec2=bvec2,
    bvec3=bvec3,
    bvec4=bvec4,
    mat2=mat2,
    mat3=mat3,
    mat2x2=mat2x2,
    mat3x2=mat3x2,
    mat4x2=mat4x2,
    mat2x3=mat2x3,
    mat3x3=mat3x3,
    mat4x3=mat4x3,
    mat2x4=mat2x4,
    mat3x4=mat3x4,
    mat4x4=mat4x4,
)
_subtypes.update(convenience_types)


# %% How to expose it all

# Types that can be referenced by name.
gpu_types_map = {}
gpu_types_map.update(leaf_types)
gpu_types_map.update(base_types)  # Only the last level, e,g. not ShaderType
gpu_types_map.update(convenience_types)


# %% Shader resource enum

RES_INPUT = "input"
RES_OUTPUT = "output"
RES_UNIFORM = "uniform"
RES_BUFFER = "buffer"
RES_SAMPLER = "sampler"
RES_TEXTURE = "texture"



"""
Implements the base class for generating SpirV code.

Notes on exceptions for code that parses/converts/compiles shading code:

* Raise ShaderError when the user-provided code cannot be compiled.
* Raise RuntimeError when the parser/compiler is in an unexpected state.
* Raise NotImplementedError only to indicate that it's an abstract method.
* Always provide an error message.
"""

import io
import struct

from ._coreutils import ShaderError
from . import _spirv_constants as cc
from . import _types


def str_to_words(s):
    # In SpirV, words are 32bit. Op counting is per word, not per immediate or per byte.
    b = s.encode()
    padding = 4 - (len(b) % 4)  # 4, 3, 2 or 1 -> always at least 1 for 0-termination
    b += padding * b"\x00"
    assert len(b) % 4 == 0 and b[-1] == 0, b
    words = []
    for i in range(0, len(b), 4):
        words.append(b[i : i + 4])
        # words.append(struct.unpack("<I", b[i : i + 4])[0])
    return words


class AnyId:
    """ Anything that has an id in a SpirV module
    """

    def __init__(self, name=""):
        self.name = name
        self.id = None

    def __repr__(self):
        clsname = self.__class__.__name__
        id = self.id or "?"

        if self.name:
            return f"<{clsname} %{self.name} ({id})>"
        else:
            return f"<{clsname} %{id}>"

    @property
    def display_name(self):
        if self.name:
            return f"%{self.name}"
        else:
            return f"%{self.id or '?'}"

    def resolve(self, gen):
        if self.id is None:
            self.id = len(gen._ids)
            gen._ids[self.id] = self
        return self


class TypeId(AnyId):
    """ A type in a SpirV module.
    """

    def __init__(self, type, name=""):
        super().__init__(name=name)
        self.type = type


class ValueId(AnyId):
    """ Anything that represents a concrete value in a SpirV module.
    """

    def __init__(self, type, name=""):
        super().__init__(name=name)
        self.type = type


class VariableAccessId(ValueId):
    """ A chain of access into a SpirV variable. The type arg is the type
    for the eventual value.
    """

    def __init__(self, variable, storage_class, type, *indices, name=""):
        super().__init__(type, name=name)
        self.variable = variable  # ValueId representing the SpirV Variable
        self.storage_class = storage_class
        self.indices = indices  # ValueId's

    def index(self, index, field=None):
        """ Index into the variable chain, so we go one level deeper.
        """
        assert isinstance(index, ValueId)
        assert issubclass(index.type, _types.Int)
        indices = list(self.indices) + [index]
        if issubclass(self.type, _types.Struct):
            assert isinstance(field, int)
            return VariableAccessId(
                self.variable,
                self.storage_class,
                self.type.get_subtype(field),
                *indices,
            )
        elif issubclass(self.type, _types.Array):
            assert field is None
            return VariableAccessId(
                self.variable, self.storage_class, self.type.subtype, *indices
            )
        elif issubclass(self.type, _types.Vector):
            assert field is None
            return VariableAccessId(
                self.variable, self.storage_class, self.type.subtype, *indices
            )
        else:
            raise ShaderError(f"VariableAccessId cannot index into {self.type}")

    def resolve_chain(self, gen):
        """ Generate OpAccessChain instruction and return pointer id object for result.
        """
        if len(self.indices) == 0:
            return self.variable
        else:
            result_type_id = gen.obtain_type_id(self.type)
            pointer_id = gen.obtain_id("pointer")
            gen.gen_instruction(
                "types",
                cc.OpTypePointer,
                pointer_id,
                self.storage_class,
                result_type_id,
            )
            result_id = gen.obtain_id("access-chain-result")
            gen.gen_func_instruction(
                cc.OpAccessChain, pointer_id, result_id, self.variable, *self.indices
            )
        return result_id

    def resolve_load(self, gen):
        """ Generate OpAccessChain instruction followed by OpLoad and return result id.
        """
        temp_id = self.resolve_chain(gen)
        id, type_id = gen.obtain_value(self.type)
        gen.gen_func_instruction(cc.OpLoad, type_id, id, temp_id)
        return id

    def resolve_store(self, gen, val):
        """ Generate OpAccessChain instruction followed by OpStore.
        """
        temp_id = self.resolve_chain(gen)
        gen.gen_func_instruction(cc.OpStore, temp_id, val)
        return val

    # Default resolve is a load
    resolve = resolve_load


class WordPlaceholder:
    """ Object that holds an integer value (or 4 bytes), which value
    can be changed as more of the code is parsed. This e.g. allows
    specifying types with knowledge that we encounter later in the
    program.
    """

    def __init__(self, initial_value):
        assert isinstance(initial_value, (int, bytes))
        self.value = initial_value

    def __repr__(self):
        return f"~{self.value}"


class BaseSpirVGenerator:
    """ Base class that can be used by compiler implementations in the
    last compile step to generate the SpirV code. It has an internal
    representation of SpirV module and provides an API to generate
    instructions. This class it not aware of our bytecode representation.
    """

    def convert(self, input):
        """ Generate the Spir-V code. After this, dump() can be used to
        produce the binary blob that represents the Spir-V module.
        """

        # Start clean
        self._init()

        # Do the thing!
        self._convert(input)

        # Wrap up
        self._post_convert()

    def _convert(self, input):
        """ Subclasses should implement this.
        """
        raise NotImplementedError()  # noqa

    def _init(self):

        self._ids = {0: None}  # maps id -> info. For objects, info is a type in _types
        self._constants = {}
        self._type_hash_to_id = {}
        self._capabilities = set()
        self._execution_modes = {}
        self._extentded_instruction_sets = {}

        # Section 2.4 of the Spir-V spec specifies the Logical Layout of a Module
        self._sections = {
            "capabilities": [],  # 1. All OpCapability instructions.
            "extensions": [],  # 2. Optional OpExtension instructions.
            "extension_imports": [],  # 3. Optional OpExtInstImport instructions.
            "memory_model": [],  # 4. The single required OpMemoryModel instruction.
            "entry_points": [],  # 5. All entry point declarations, using OpEntryPoint.
            "execution_modes": [],  # 6. All execution-mode declarations, using OpExecutionMode or OpExecutionModeId.
            "debug": [],  # 7. The debug instructions, which must be grouped in a specific following order.
            "annotations": [],  # 8. All annotation instructions, e.g. OpDecorate.
            "types": [],  # 9. All type declarations (OpTypeXXX instructions),
            # all constant instructions, and all global
            # variable declarations (all OpVariable instructions whose
            # Storage Class is notFunction). This is the preferred
            # location for OpUndef instructions, though they can also
            # appear in function bodies. All operands in all these
            # instructions must be declared before being used. Otherwise,
            # they can be in any order. This section is the ﬁrst section
            # to allow use of OpLine debug information.
            "function_defs": [],  # 10. All function declarations. A function
            # declaration is as follows.
            # a. Function declaration, using OpFunction.
            # b. Function parameter declarations, using OpFunctionParameter.
            # c. Function end, using OpFunctionEnd.
            "functions": [],  # 11. All function deﬁnitions (functions with a body).
            # A function deﬁnition is as follows:
            # a. Function deﬁnition, using OpFunction.
            # b. Function parameter declarations, using OpFunctionParameter.
            # c. Block, Block ...
            # d. Function end, using OpFunctionEnd.
        }

    def _post_convert(self):
        """ After most of the generation has been done, we set the required capabilities
        and massage the order of instructions a bit.
        """

        # Define memory model (1 instruction)
        self.gen_instruction(
            "memory_model",
            cc.OpMemoryModel,
            cc.AddressingModel_Logical,
            cc.MemoryModel_Simple,
        )

        # Write execution modes
        for mode_name, mode_args in self._execution_modes.items():
            self.gen_instruction(
                "execution_modes",
                cc.OpExecutionMode,
                self._entry_point_id,
                getattr(cc, "ExecutionMode_" + mode_name),
                *mode_args,
            )

        # Remove duplicate types. This is required because some types are not
        # "complete" until the shader has been fully parsed. In particular the
        # OpTypeImage.
        type_instructions = self._sections["types"]
        seen_type_defs = {}
        to_remove = []
        for i in range(len(type_instructions)):
            words = type_instructions[i]
            if words[0] not in (cc.OpTypeImage,):
                continue
            # Resolve WordPlaceholder's
            words = tuple(
                w.value if isinstance(w, WordPlaceholder) else w for w in words
            )
            type_instructions[i] = words
            # Get hash of the parts except the TypeId itself, see if we already have it.
            # If so, replace id with the id of the other, and remove from list.
            h = hash(words[:1] + words[2:])
            if h in seen_type_defs:
                words[1].id = seen_type_defs[h][1].id
                to_remove.append(i)
            else:
                seen_type_defs[h] = words
        for i in reversed(to_remove):
            type_instructions.pop(i)

        # Define capabilities. We "collect" these as certain features are used.
        self.gen_instruction("capabilities", cc.OpCapability, cc.Capability_Shader)
        for capability_op in sorted(self._capabilities):
            self.gen_instruction("capabilities", cc.OpCapability, capability_op)

        # Move OpVariable to the start of a function
        # Variables are used to refer to either internal variables, or IO, and load/store
        # is used to move variables from/to the stack.
        func_instructions = self._sections["functions"]
        insert_point = -1
        for i in range(len(func_instructions)):
            if func_instructions[i][0] == cc.OpFunction:
                insert_point = i + 2
            elif func_instructions[i][0] == cc.OpVariable:
                func_instructions.insert(insert_point, func_instructions.pop(i))
                insert_point += 1

        # Get ids of global variables
        global_OpVariable_s = []
        for instr in self._sections["types"]:
            if instr[0] == cc.OpVariable:
                # todo: can remove the if below when we move to 1.4 or above
                if instr[-1] in (cc.StorageClass_Input, cc.StorageClass_Output):
                    global_OpVariable_s.append(instr[2])
        # We assume one function, so all are used in our single function
        self._sections["entry_points"][0] = self._sections["entry_points"][0] + tuple(
            global_OpVariable_s
        )

    # %% Utility for compiler

    def to_text(self):
        """ Generate a textual (dis-assembly-like) representation.
        """

        lines = []
        edge = 22

        def disp(pre, pro):
            pre = pre or ""
            line = str(pre.rjust(edge)) + str(pro)
            lines.append(line)

        disp("header ".ljust(edge, "-"), "")
        disp("MagicNumber: ", hex(cc.MagicNumber))
        disp("Version: ", hex(cc.Version))
        disp("VendorId: ", hex(0))
        disp("Bounds: ", len(self._ids))
        disp("Reserved: ", hex(0))

        seen_ids = set()
        for section_name, instructions in self._sections.items():
            # disp(section_name.upper(), "-" * 20)
            disp((section_name + " ").ljust(edge, "-"), "")
            for instruction in instructions:
                instruction_str = repr(instruction[0])
                ret = None
                for i in instruction[1:]:
                    if isinstance(i, AnyId):
                        i_str = i.display_name
                        if instruction[0] in (
                            cc.OpDecorate,
                            cc.OpLoopMerge,
                            cc.OpSelectionMerge,
                            cc.OpBranch,
                            cc.OpBranchConditional,
                        ):
                            pass
                        elif i.id not in seen_ids:
                            seen_ids.add(i.id)
                            ret = i.display_name + " = "
                            i_str = f"({i.id})"
                        instruction_str += " " + i_str
                    else:
                        instruction_str += " " + repr(i)
                disp(ret, instruction_str)

        return "\n".join(lines)

    def dump(self):
        """ Generated a bytes object representing the Spir-V module.
        """

        f = io.BytesIO()

        def write_word(w):
            if isinstance(w, bytes):
                assert len(w) == 4
                f.write(w)
            else:
                f.write(struct.pack("<I", w))

        # We pin to version 1.3, since higher versions seem not well supported by drivers
        # version = cc.Version
        # version = struct.unpack("<I", struct.pack("<bbbb", 0, 5, 1, 0))
        version = 66304

        # Write header
        write_word(cc.MagicNumber)  # Magic number
        write_word(version)  # SpirV version
        write_word(0)  # Vendor id - can be zero, let's use zero until we are registered
        write_word(len(self._ids))  # Bound (of ids)
        write_word(0)  # Reserved

        # Write instructions
        for instructions in self._sections.values():
            for opcode, *instr_words in instructions:
                words = []
                for word in instr_words:
                    if isinstance(word, AnyId):
                        words.append(word.id)
                    elif isinstance(word, WordPlaceholder):
                        words.append(word.value)
                    elif isinstance(word, str):
                        words.extend(str_to_words(word))
                    else:
                        words.append(word)
                write_word(((len(words) + 1) << 16) | opcode)
                for word in words:
                    write_word(word)

        return f.getvalue()

    # %% Utils for subclasses

    def gen_instruction(self, section_name, opcode, *words):
        words = [
            word.resolve(self) if isinstance(word, AnyId) else word for word in words
        ]
        self._sections[section_name].append((opcode, *words))

    def gen_func_instruction(self, opcode, *words):
        words = [
            word.resolve(self) if isinstance(word, AnyId) else word for word in words
        ]
        self._sections["functions"].append((opcode, *words))

    def obtain_id(self, name=""):
        """ Get a new raw id for anything that's not a value or type.
        """
        return AnyId(name=name)

    def obtain_value(self, the_type, name=""):
        """ Create id for a new value. Returns (value_id, type_id).
        """
        type_id = self.obtain_type_id(the_type)
        value_id = ValueId(the_type)
        return value_id, type_id
        # todo: return only value, and support value.type_id?

    def obtain_constant(self, value, the_type=None):
        """ Get the id object for the constant of given value.
        Existing constants are re-used.
        """
        # First derive SpirV type from value
        if isinstance(value, float):
            the_type = _types.f32 if the_type is None else the_type
            assert the_type.__name__ != "f16", "Cannot yet create f16 constants."
            M = {"f32": "<f", "f64": "<d"}
            struct_type = M[the_type.__name__]
            bb = struct.pack(struct_type, value)
        elif isinstance(value, bool):  # test before int because issubclass(bool, int)
            the_type = _types.boolean
        elif isinstance(value, int):
            the_type = _types.i32 if the_type is None else the_type
            M = {"u8": "<B", "i16": "<h", "i32": "<i", "i64": "<q"}
            struct_type = M[the_type.__name__]
            bb = struct.pack(struct_type, value)
        else:
            raise RuntimeError(f"Cannot get a constant for {value}")
        # Make sure that we have it
        key = the_type.__name__, value
        if key not in self._constants:
            id, type_id = self.obtain_value(the_type)
            if the_type is _types.boolean:
                opcode = cc.OpConstantTrue if value else cc.OpConstantFalse
                self.gen_instruction("types", opcode, type_id, id)
            else:
                self.gen_instruction("types", cc.OpConstant, type_id, id, bb)
            self._constants[key] = id
        # Return cached
        return self._constants[key]

    def obtain_variable(self, the_type, storage_class, name=""):
        """ Create a variable in the current scope. Generates an OpVariable
        definition instruction and returns a VariableAccessId to access it.
        """
        # Create id and type_id
        var_id, var_type_id = self.obtain_value(the_type, name)
        #  Create pointer for variable
        var_pointer_id = self.obtain_id("pointer")
        self.gen_instruction(
            "types", cc.OpTypePointer, var_pointer_id, storage_class, var_type_id
        )
        # Generate the variable instruction
        where = "types"
        if storage_class in (cc.StorageClass_Function, cc.StorageClass_Private):
            where = "functions"
        self.gen_instruction(
            where, cc.OpVariable, var_pointer_id, var_id, storage_class
        )
        # Return object that can be used to access the variable with multi-level indexing
        return VariableAccessId(var_id, storage_class, the_type, name=name)

    def obtain_type_id(self, the_type):
        """ Get the id for the given type. Generates a type
        definition instruction as needed.
        """
        if isinstance(the_type, tuple):
            if not (
                the_type
                and isinstance(the_type[0], cc.Enum)
                and the_type[0].name.startswith("OpType")
            ):
                raise RuntimeError(
                    "ShaderType can be tuple only if it specifies the OpTypeXYZ"
                )
            type_hash = hash(the_type)
        else:
            if not (
                isinstance(the_type, type) and issubclass(the_type, _types.ShaderType)
            ):
                raise RuntimeError(f"not a ShaderType subclass: {the_type}")
            assert not the_type.is_abstract, f"not a concrete spirv type: {the_type}"
            type_hash = the_type.__name__

        # Already know this type?
        if type_hash in self._type_hash_to_id:
            return self._type_hash_to_id[type_hash]

        if isinstance(the_type, tuple):
            type_id = TypeId(the_type)  # all info is now on TypeId instance
            self.gen_instruction("types", the_type[0], type_id, *the_type[1:])
        elif issubclass(the_type, _types.void):
            type_id = TypeId(the_type)
            self.gen_instruction("types", cc.OpTypeVoid, type_id)
        elif issubclass(the_type, _types.boolean):
            type_id = TypeId(the_type)
            self.gen_instruction("types", cc.OpTypeBool, type_id)
        elif issubclass(the_type, _types.Int):
            type_id = TypeId(the_type)
            if issubclass(the_type, _types.u8):
                self._capabilities.add(cc.Capability_Int8)
                self.gen_instruction("types", cc.OpTypeInt, type_id, 8, 0)
            elif issubclass(the_type, _types.i16):
                self._capabilities.add(cc.Capability_Int16)
                self.gen_instruction("types", cc.OpTypeInt, type_id, 16, 1)
            elif issubclass(the_type, _types.i32):
                self.gen_instruction("types", cc.OpTypeInt, type_id, 32, 1)
            elif issubclass(the_type, _types.i64):
                self._capabilities.add(cc.Capability_Int64)
                self.gen_instruction("types", cc.OpTypeInt, type_id, 64, 1)
            else:
                raise RuntimeError(f"Unknown integer type: {the_type}")
        elif issubclass(the_type, _types.Float):
            type_id = TypeId(the_type)
            if issubclass(the_type, _types.f16):
                self._capabilities.add(cc.Capability_Float16)
                self.gen_instruction("types", cc.OpTypeFloat, type_id, 16)
            elif issubclass(the_type, _types.f32):
                self.gen_instruction("types", cc.OpTypeFloat, type_id, 32)
            elif issubclass(the_type, _types.f64):
                self._capabilities.add(cc.Capability_Float64)
                self.gen_instruction("types", cc.OpTypeFloat, type_id, 64)
            else:
                raise RuntimeError(f"Unknown float type: {the_type}")
        elif issubclass(the_type, _types.Vector):
            sub_type_id = self.obtain_type_id(the_type.subtype)
            type_id = TypeId(the_type)
            self.gen_instruction(
                "types", cc.OpTypeVector, type_id, sub_type_id, the_type.length
            )
        elif issubclass(the_type, _types.Matrix):
            sub_vector_type = _types.Vector(the_type.rows, the_type.subtype)
            column_type_id = self.obtain_type_id(sub_vector_type)
            type_id = TypeId(the_type)
            self.gen_instruction(
                "types", cc.OpTypeMatrix, type_id, column_type_id, the_type.cols
            )
        elif issubclass(the_type, _types.Array):
            count = the_type.length
            sub_type_id = self.obtain_type_id(the_type.subtype)
            if not count:
                # An array for which the length is not known at compile time
                # Use OpArrayLength to get the array length at compile time
                type_id = TypeId(the_type)
                self.gen_instruction(
                    "types", cc.OpTypeRuntimeArray, type_id, sub_type_id
                )
            else:
                # Handle count
                count_value_id = self.obtain_constant(count)
                # Handle toplevel array type
                type_id = TypeId(the_type)
                self.gen_instruction(
                    "types", cc.OpTypeArray, type_id, sub_type_id, count_value_id
                )
        elif issubclass(the_type, _types.Struct):
            type_id = TypeId(the_type)
            subtype_ids = [
                self.obtain_type_id(the_type.get_subtype(key)) for key in the_type.keys
            ]
            self.gen_instruction("types", cc.OpTypeStruct, type_id, *subtype_ids)
        else:
            raise RuntimeError(f"Unknown GPU type {the_type}")

        self._type_hash_to_id[type_hash] = type_id
        return type_id

    def obtain_extended_instruction_set(self, set_name):
        """ Obtain the extended instruction set object by the instruction set name.
        The used instruction sets are defined near the top of the SpirV file. The
        resulting id is used in OpExtInst instructions.
        """
        if set_name not in self._extentded_instruction_sets:
            id = self.obtain_id(set_name)
            self.gen_instruction("extension_imports", cc.OpExtInstImport, id, set_name)
            self._extentded_instruction_sets[set_name] = id
        return self._extentded_instruction_sets[set_name]



from setuptools import find_packages, setup


def get_version_and_docstring():
    ns = {"__doc__": "", "__version__": ""}
    docStatus = 0  # Not started, in progress, done
    for line in open("pyshader/__init__.py").readlines():
        if line.startswith("__version__"):
            exec(line.strip(), ns, ns)
        elif line.startswith('"""'):
            if docStatus == 0:
                docStatus = 1
                line = line.lstrip('"')
            elif docStatus == 1:
                docStatus = 2
        if docStatus == 1:
            ns["__doc__"] += line.rstrip() + "\n"
    return ns["__version__"], ns["__doc__"]


version, doc = get_version_and_docstring()

setup(
    name="pyshader",
    version=version,
    url="https://github.com/pygfx/pyshader",
    description="Write modern GPU shaders in Python!",
    long_description=doc,
    long_description_content_type="text/markdown",
    author="Almar Klein",
    author_email="almar.klein@gmail.com",
    packages=find_packages(
        exclude=["tests", "tests.*", "examples_py", "examples_py.*"]
    ),
    python_requires=">=3.6.0",
    zip_safe=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)
